from __future__ import annotations

import datetime
import itertools
import json
from abc import abstractmethod

from ravendb.documents.operations.executor import OperationExecutor
from ravendb.documents.session.document_session_operations.misc import _update_metadata_modifications
from ravendb.documents.operations.time_series import TimeSeriesRangeResult
from ravendb.documents.session.misc import (
    SessionOptions,
    TransactionMode,
    SessionInfo,
    ForceRevisionStrategy,
    DocumentsChanges,
)
from ravendb.tools.time_series import TSRangeHelper

try:
    from collections.abc import MutableSet
except Exception:
    from collections import MutableSet

import uuid as uuid
from copy import deepcopy, Error
from typing import Union, Callable, List, Dict, Set, Type, TypeVar, Tuple, Any

from ravendb.primitives import constants
from ravendb.documents.commands.crud import GetDocumentsResult
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.identity.hilo import GenerateEntityIdOnTheClient
from ravendb.exceptions.exceptions import NonUniqueObjectException, InvalidOperationException
from ravendb.documents.commands.batches import (
    CommandData,
    BatchOptions,
    DeleteCommandData,
    CommandType,
    BatchPatchCommandData,
    PutCommandDataWithJson,
    ForceRevisionCommandData,
    IndexBatchOptions,
    ReplicationBatchOptions,
)
from ravendb.documents.session.cluster_transaction_operation import ClusterTransactionOperationsBase
from ravendb.documents.session.concurrency_check_mode import ConcurrencyCheckMode
from ravendb.documents.session.document_info import DocumentInfo
from ravendb.documents.session.event_args import *
from ravendb.documents.session.time_series import TimeSeriesEntry
from ravendb.documents.session.utils.includes_util import IncludesUtil
from ravendb.extensions.json_extensions import JsonExtensions
from ravendb.http.raven_command import RavenCommand
from ravendb.json.json_operation import JsonOperation
from ravendb.json.result import BatchCommandResult
from ravendb.documents.session.entity_to_json import EntityToJson
from ravendb.tools.utils import Utils, CaseInsensitiveDict, CaseInsensitiveSet
from ravendb.documents.store.misc import IdTypeAndName

if TYPE_CHECKING:
    from ravendb.documents.operations.lazy.definition import LazyOperation
    from ravendb.documents.store.definition import DocumentStore
    from ravendb.http.request_executor import RequestExecutor


class RefEq:
    ref = None

    def __init__(self, ref):
        if isinstance(ref, RefEq):
            self.ref = ref.ref
            return
        self.ref = ref

    # As we split the hashable and unhashable items into separate collections, we only compare _RefEq to other _RefEq
    def __eq__(self, other):
        if isinstance(other, RefEq):
            return id(self.ref) == id(other.ref)
        raise TypeError("Expected _RefEq type object")

    def __hash__(self):
        return id(self.ref)


class RefEqEntityHolder(object):
    def __init__(self):
        self.__unhashable_items = dict()

    def __len__(self):
        return len(self.__unhashable_items)

    def __contains__(self, item):
        return RefEq(item) in self.__unhashable_items

    def __delitem__(self, key):
        del self.__unhashable_items[RefEq(key)]

    def __setitem__(self, key, value):
        self.__unhashable_items[RefEq(key)] = value

    def __getitem__(self, key):
        return self.__unhashable_items[RefEq(key)].ref

    def __getattribute__(self, item):
        if item == "_RefEqEntityHolder__unhashable_items":
            return super().__getattribute__(item)
        return self.__unhashable_items.__getattribute__(item)


class DocumentsByIdHolder(object):
    def __init__(self):
        self.__inner = CaseInsensitiveDict()

    def __len__(self):
        return len(self.__inner)

    def __iter__(self):
        return self.__inner.__iter__()

    def __delitem__(self, key):
        del self.__inner[key]

    def __setitem__(self, key, value):
        self.__inner[key] = value

    def __getattribute__(self, item):
        if item in ["_DocumentsByIdHolder__inner", "get_value", "add", "remove", "clear"]:
            return super().__getattribute__(item)
        return self.__inner.__getattribute__(item)

    def get_value(self, key: str) -> DocumentInfo:
        return self.__inner.get(key)

    def add(self, info: DocumentInfo) -> None:
        if info.key in self.__inner:
            return

        self.__inner[info.key] = info

    def remove(self, key: str) -> bool:
        return self.__inner.pop(key, None) is not None

    def clear(self) -> None:
        self.__inner.clear()


class DocumentsByEntityHolder(object):
    def __init__(self):
        self.__documents_by_entity_hashable: Dict[object, DocumentInfo] = dict()
        self.__documents_by_entity_unhashable: RefEqEntityHolder[RefEq, DocumentInfo] = RefEqEntityHolder()
        self.__on_before_store_documents_by_entity_hashable: Dict[object, DocumentInfo] = dict()
        self.__on_before_store_documents_by_entity_unhashable: RefEqEntityHolder[RefEq, DocumentInfo] = (
            RefEqEntityHolder()
        )
        self.__prepare_entities_puts: bool = False

    def __repr__(self):
        return f"{self.__class__.__name__}: {[item for item in self]}"

    def __len__(self):
        return (
            len(self.__documents_by_entity_hashable)
            + len(self.__documents_by_entity_unhashable)
            + (
                len(self.__on_before_store_documents_by_entity_hashable)
                if self.__on_before_store_documents_by_entity_hashable
                else 0
            )
            + (
                len(self.__on_before_store_documents_by_entity_unhashable)
                if self.__on_before_store_documents_by_entity_unhashable
                else 0
            )
        )

    def __contains__(self, item):
        if not self.__prepare_entities_puts:
            try:
                return item in self.__documents_by_entity_hashable
            except TypeError as e:
                if str(e.args[0]).startswith("unhashable type"):
                    return item in self.__documents_by_entity_unhashable
                raise e
        self.__create_on_before_store_documents_by_entity_if_needed()
        try:
            return item in self.__on_before_store_documents_by_entity_hashable
        except TypeError as e:
            if str(e.args[0]).startswith("unhashable type"):
                return item in self.__documents_by_entity_unhashable
            raise e

    def __setitem__(self, key, value):
        if not self.__prepare_entities_puts:
            try:
                self.__documents_by_entity_hashable[key] = value
            except TypeError as e:
                if str(e.args[0]).startswith("unhashable type"):
                    self.__documents_by_entity_unhashable[key] = value
                    return
                raise e
            return
        self.__create_on_before_store_documents_by_entity_if_needed()
        try:
            self.__on_before_store_documents_by_entity_hashable[key] = value
        except TypeError as e:
            if str(e.args[0]).startswith("unhashable type"):
                self.__on_before_store_documents_by_entity_unhashable[key] = value
                return
            raise e

    def __create_on_before_store_documents_by_entity_if_needed(self) -> None:
        if not self.__on_before_store_documents_by_entity_unhashable:
            self.__on_before_store_documents_by_entity_unhashable = RefEqEntityHolder()
        if not self.__on_before_store_documents_by_entity_hashable:
            self.__on_before_store_documents_by_entity_hashable = dict()
        return

    def __getitem__(self, key):
        try:
            document_info = self.__documents_by_entity_hashable[key]
        except (TypeError, KeyError):
            document_info = self.__documents_by_entity_unhashable.get(RefEq(key), None)
        if document_info:
            return document_info
        if self.__on_before_store_documents_by_entity_hashable:
            document_info = self.__on_before_store_documents_by_entity_hashable.get(key, None)
        if not document_info and self.__documents_by_entity_unhashable:
            return self.__on_before_store_documents_by_entity_unhashable.get(RefEq(key), None)
        return None

    def __iter__(self):
        first_hashable = (
            (
                self.DocumentsByEntityEnumeratorResult(key, value, True)
                for key, value in self.__documents_by_entity_hashable.items()
            )
            if len(self.__documents_by_entity_hashable) > 0
            else None
        )

        first_unhashable = (
            (
                self.DocumentsByEntityEnumeratorResult(key, value, True)
                for key, value in self.__documents_by_entity_unhashable.items()
            )
            if len(self.__documents_by_entity_unhashable) > 0
            else None
        )

        second_hashable = (
            (
                self.DocumentsByEntityEnumeratorResult(key, value, False)
                for key, value in self.__on_before_store_documents_by_entity_hashable.items()
            )
            if len(self.__on_before_store_documents_by_entity_hashable) > 0
            else None
        )
        second_unhashable = (
            (
                self.DocumentsByEntityEnumeratorResult(key, value, False)
                for key, value in self.__on_before_store_documents_by_entity_unhashable.items()
            )
            if len(self.__on_before_store_documents_by_entity_unhashable) > 0
            else None
        )

        return itertools.chain(
            *[x for x in [first_hashable, first_unhashable, second_hashable, second_unhashable] if x]
        )

    def get(self, key, default=None):
        return self[key] if key in self else default

    def pop(self, key, default_value=None):
        result = self.__documents_by_entity_hashable.pop(key, None)
        if not result:
            result = self.__documents_by_entity_unhashable.pop(RefEq(key), default_value)

        if self.__on_before_store_documents_by_entity_unhashable:
            if not result:
                result = self.__on_before_store_documents_by_entity_unhashable.pop(RefEq(key), None)
            else:
                del self.__on_before_store_documents_by_entity_unhashable[RefEq(key)]
        if self.__on_before_store_documents_by_entity_hashable:
            if not result:
                result = self.__on_before_store_documents_by_entity_hashable.pop(key, None)
            else:
                del self.__on_before_store_documents_by_entity_hashable

        return result

    def evict(self, entity: object) -> None:
        if self.__prepare_entities_puts:
            raise RuntimeError("Cannot evict entity during on_before_store")
        result = self.__documents_by_entity_unhashable.pop(RefEq(entity), None)
        if result is None:
            self.__documents_by_entity_hashable.pop(entity, None)

    def clear(self):
        self.__documents_by_entity_hashable.clear()
        self.__documents_by_entity_unhashable.clear()
        if self.__on_before_store_documents_by_entity_hashable:
            self.__on_before_store_documents_by_entity_hashable.clear()
        if self.__on_before_store_documents_by_entity_unhashable:
            self.__on_before_store_documents_by_entity_unhashable.clear()

    class DocumentsByEntityEnumeratorResult:
        def __init__(self, key: object, value: DocumentInfo, execute_on_before_store: bool):
            self.__key = key.ref if isinstance(key, RefEq) else key
            self.__value = value
            self.__execute_on_before_store = execute_on_before_store

        @property
        def key(self):
            return self.__key

        @property
        def value(self):
            return self.__value

        @property
        def execute_on_before_store(self):
            return self.__execute_on_before_store


class DeletedEntitiesHolder(MutableSet):
    def __init__(self, items=None, prepare_entities_deletes: bool = False):
        if items is None:
            items = []
        self.__deleted_entities = set(map(RefEq, items))
        self.__prepare_entities_deleted = prepare_entities_deletes
        self.__on_before_deleted_entities: Union[set, None] = None

    def __getattribute__(self, item):
        if item in [
            "add",
            "remove",
            "evict",
            "clear",
            "discard",
            "DeletedEntitiesEnumeratorResult",
            "_DeletedEntitiesHolder__deleted_entities",
            "_DeletedEntitiesHolder__prepare_entities_deleted",
            "_DeletedEntitiesHolder__on_before_deleted_entities",
        ]:
            return super().__getattribute__(item)
        return self.__deleted_entities.__getattribute__(item)  # todo: getattribute of joined sets, not just one..

    def __contains__(self, item: object) -> bool:
        if RefEq(item) in self.__deleted_entities:
            return True
        if self.__on_before_deleted_entities is None:
            return False
        return RefEq(item) in self.__on_before_deleted_entities

    def __len__(self) -> int:
        return len(self.__deleted_entities) + (
            len(self.__on_before_deleted_entities) if self.__on_before_deleted_entities else 0
        )

    def __iter__(self):
        deleted_transformed_iterator = (
            self.DeletedEntitiesEnumeratorResult(item.ref, True) for item in self.__deleted_entities
        )
        if self.__on_before_deleted_entities is None:
            return deleted_transformed_iterator

        on_before_deleted_iterator = (
            self.DeletedEntitiesEnumeratorResult(item.ref, False) for item in self.__on_before_deleted_entities
        )
        return itertools.chain(deleted_transformed_iterator, on_before_deleted_iterator)

    def add(self, element: object) -> None:
        if self.__prepare_entities_deleted:
            if self.__on_before_deleted_entities is None:
                self.__on_before_deleted_entities = set()
            self.__on_before_deleted_entities.add(RefEq(element))
            return
        return self.__deleted_entities.add(RefEq(element))

    def remove(self, entity: object) -> None:
        self.__deleted_entities.remove(RefEq(entity))
        if self.__on_before_deleted_entities:
            self.__on_before_deleted_entities.remove(RefEq(entity))

    def discard(self, element: object) -> None:
        self.__deleted_entities.discard(RefEq(element))
        if self.__on_before_deleted_entities is not None:
            self.__on_before_deleted_entities.remove(RefEq(element))

    def clear(self) -> None:
        self.__deleted_entities.clear()
        if self.__on_before_deleted_entities:
            self.__on_before_deleted_entities.clear()

    def evict(self, entity) -> None:
        if self.__prepare_entities_deleted:
            raise RuntimeError("Cannot evict entity during OnBeforeDelete")
        self.__deleted_entities.discard(RefEq(entity))

    class DeletedEntitiesEnumeratorResult:
        def __init__(self, entity: object, execute_on_before_delete: bool):
            self.__entity = entity
            self.__execute_on_before_delete = execute_on_before_delete

        @property
        def entity(self) -> object:
            return self.__entity

        @property
        def execute_on_before_delete(self) -> bool:
            return self.__execute_on_before_delete


_T = TypeVar("_T")


class InMemoryDocumentSessionOperations:
    __instances_counter: int = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __init__(self, store: "DocumentStore", key: uuid.UUID, options: SessionOptions):
        self._id = key
        self._database_name = options.database if options.database else store.database if store.database else None
        if not self._database_name:
            InMemoryDocumentSessionOperations.raise_no_database()

        self._request_executor = (
            options.request_executor if options.request_executor else store.get_request_executor(self._database_name)
        )
        self._operation_executor: Union[None, OperationExecutor] = None

        self._pending_lazy_operations: List[LazyOperation] = []
        self._on_evaluate_lazy = {}

        self._no_tracking = options.no_tracking

        self._use_optimistic_concurrency = self._request_executor.conventions.use_optimistic_concurrency
        self._max_number_of_requests_per_session = self._request_executor.conventions.max_number_of_requests_per_session
        self._generate_entity_id_on_client = GenerateEntityIdOnTheClient(
            self._request_executor.conventions, self._generate_id
        )
        self.entity_to_json = EntityToJson(self)
        self._document_store: DocumentStore = store
        self.session_info = SessionInfo(self, options, self._document_store)
        self._save_changes_options = BatchOptions()

        self.transaction_mode = options.transaction_mode
        self.disable_atomic_document_writes_in_cluster_wide_transaction = (
            options.disable_atomic_document_writes_in_cluster_wide_transaction
        )

        self._known_missing_ids = CaseInsensitiveSet()
        self._documents_by_id = DocumentsByIdHolder()
        self._included_documents_by_id = CaseInsensitiveDict()
        self.include_revisions_by_change_vector = CaseInsensitiveDict()
        self.include_revisions_by_date_time_before: Optional[Dict[str, Dict[datetime.datetime, DocumentInfo]]] = (
            CaseInsensitiveDict()
        )
        self._documents_by_entity: DocumentsByEntityHolder = DocumentsByEntityHolder()

        self._counters_by_doc_id: Dict[str, List[Dict[str, int]]] = CaseInsensitiveDict()
        self._time_series_by_doc_id: Dict[str, Dict[str, List[TimeSeriesRangeResult]]] = CaseInsensitiveDict()

        self._deleted_entities: Union[
            Set[DeletedEntitiesHolder.DeletedEntitiesEnumeratorResult], DeletedEntitiesHolder
        ] = DeletedEntitiesHolder()
        self._deferred_commands: List[CommandData] = []
        self._deferred_commands_map: Dict[IdTypeAndName, CommandData] = {}
        self._ids_for_creating_forced_revisions: Dict[str, ForceRevisionStrategy] = CaseInsensitiveDict()
        # todo: pendingLazyOperations, onEvaluateLazy
        self._generate_document_keys_on_store: bool = True
        self._save_changes_options: Union[None, BatchOptions] = None
        self._hash: int = self.__instances_counter.__add__(1)
        self._is_disposed: Union[None, bool] = None

        self._number_of_requests: int = 0
        self._max_number_of_requests_per_session: int = (
            self._request_executor.conventions.max_number_of_requests_per_session
        )

        # --- EVENTS ---
        self._before_store: List[Callable[[BeforeStoreEventArgs], None]] = []
        self._after_save_changes: List[Callable[[AfterSaveChangesEventArgs], None]] = []
        self._before_delete: List[Callable[[BeforeDeleteEventArgs], None]] = []
        self._before_query: List[Callable[[BeforeQueryEventArgs], None]] = []
        self._before_conversion_to_document: List[Callable[[BeforeConversionToDocumentEventArgs], None]] = []
        self._after_conversion_to_document: List[Callable[[AfterConversionToDocumentEventArgs], None]] = []

        self._before_conversion_to_entity: List[Callable[[BeforeConversionToEntityEventArgs], None]] = []

        self._after_conversion_to_entity: List[Callable[[AfterConversionToEntityEventArgs], None]] = []

        self._session_closing: List[Callable[[SessionClosingEventArgs], None]] = []

    def add_before_store(self, event: Callable[[BeforeStoreEventArgs], None]):
        self._before_store.append(event)

    def remove_before_store(self, event: Callable[[BeforeStoreEventArgs], None]):
        self._before_store.remove(event)

    def add_session_closing(self, event: Callable[[SessionClosingEventArgs], None]):
        self._session_closing.append(event)

    def remove_session_closing(self, event: Callable[[SessionClosingEventArgs], None]):
        self._session_closing.remove(event)

    def add_before_conversion_to_document(self, event: Callable[[BeforeConversionToDocumentEventArgs], None]):
        self._before_conversion_to_document.append(event)

    def remove_before_conversion_to_document(self, event: Callable[[BeforeConversionToDocumentEventArgs], None]):
        self._before_conversion_to_document.remove(event)

    def add_after_conversion_to_document(self, event: Callable[[AfterConversionToDocumentEventArgs], None]):
        self._after_conversion_to_document.append(event)

    def remove_after_conversion_to_document(self, event: Callable[[AfterConversionToDocumentEventArgs], None]):
        self._after_conversion_to_document.remove(event)

    def add_before_conversion_to_entity(self, event: Callable[[BeforeConversionToEntityEventArgs], None]):
        self._before_conversion_to_entity.append(event)

    def remove_before_conversion_to_entity(self, event: Callable[[BeforeConversionToEntityEventArgs], None]):
        self._before_conversion_to_entity.remove(event)

    def add_after_conversion_to_entity(self, event: Callable[[AfterConversionToEntityEventArgs], None]):
        self._after_conversion_to_entity.append(event)

    def remove_after_conversion_to_entity(self, event: Callable[[AfterConversionToEntityEventArgs], None]):
        self._after_conversion_to_entity.remove(event)

    def add_after_save_changes(self, event: Callable[[AfterSaveChangesEventArgs], None]):
        self._after_save_changes.append(event)

    def remove_after_save_changes(self, event: Callable[[AfterSaveChangesEventArgs], None]):
        self._after_save_changes.remove(event)

    def add_before_delete(self, event: Callable[[BeforeDeleteEventArgs], None]):
        self._before_delete.append(event)

    def remove_before_delete_entity(self, event: Callable[[BeforeDeleteEventArgs], None]):
        self._before_delete.remove(event)

    def add_before_query(self, event: Callable[[BeforeQueryEventArgs], None]):
        self._before_query.append(event)

    def remove_before_query(self, event: Callable[[BeforeQueryEventArgs], None]):
        self._before_query.append(event)

    def before_store_invoke(self, before_store_event_args: BeforeStoreEventArgs):
        for event in self._before_store:
            event(before_store_event_args)

    def session_closing_invoke(self, session_closing_event_args: SessionClosingEventArgs):
        for event in self._session_closing:
            event(session_closing_event_args)

    def before_conversion_to_document_invoke(
        self, before_conversion_to_document_event_args: BeforeConversionToDocumentEventArgs
    ):
        for event in self._before_conversion_to_document:
            event(before_conversion_to_document_event_args)

    def after_conversion_to_document_invoke(
        self, after_conversion_to_document_event_args: AfterConversionToDocumentEventArgs
    ):
        for event in self._after_conversion_to_document:
            event(after_conversion_to_document_event_args)

    def before_conversion_to_entity_invoke(
        self, before_conversion_to_entity_event_args: BeforeConversionToEntityEventArgs
    ):
        for event in self._before_conversion_to_entity:
            event(before_conversion_to_entity_event_args)

    def after_conversion_to_entity_invoke(
        self, after_conversion_to_entity_event_args: AfterConversionToEntityEventArgs
    ):
        for event in self._after_conversion_to_entity:
            event(after_conversion_to_entity_event_args)

    def after_save_changes_invoke(self, after_save_changes_event_args: AfterSaveChangesEventArgs):
        for event in self._after_save_changes:
            event(after_save_changes_event_args)

    def before_delete_invoke(self, before_delete_event_args: BeforeDeleteEventArgs):
        for event in self._before_delete:
            event(before_delete_event_args)

    def before_query_invoke(self, before_query_event_args: BeforeQueryEventArgs):
        for event in self._before_query:
            event(before_query_event_args)

    @property
    def documents_by_id(self):
        return self._documents_by_id

    @property
    def deleted_entities(self):
        return self._deleted_entities

    @property
    def deferred_commands_map(self):
        return self._deferred_commands_map

    @property
    def documents_by_entity(self):
        return self._documents_by_entity

    @property
    def operations(self) -> OperationExecutor:
        return self._operation_executor

    @property
    def conventions(self) -> DocumentConventions:
        return self._request_executor.conventions

    @property
    def _store_identifier(self):
        return self._document_store.identifier

    @property
    def database_name(self):
        return self._database_name

    @property
    def generate_entity_id_on_the_client(self) -> GenerateEntityIdOnTheClient:
        return self._generate_entity_id_on_client

    @property
    def counters_by_doc_id(self):
        if self._counters_by_doc_id is None:
            self._counters_by_doc_id = CaseInsensitiveDict()
        return self._counters_by_doc_id

    @property
    def time_series_by_doc_id(self) -> Dict[str, Dict[str, List[TimeSeriesRangeResult]]]:
        return self._time_series_by_doc_id

    @property
    def number_of_requests(self) -> int:
        return self._number_of_requests

    @property
    def document_store(self) -> DocumentStore:
        return self._document_store

    @property
    def request_executor(self) -> RequestExecutor:
        return self._request_executor

    @property
    def ids_for_creating_forced_revisions(self):
        return self._ids_for_creating_forced_revisions

    @property
    def no_tracking(self) -> bool:
        return self._no_tracking

    @no_tracking.setter
    def no_tracking(self, value: bool):
        self._no_tracking = value

    # def counters_for(self, entity_or_document_id):
    #     """
    #     Get A counters object associated with the document
    #     """
    #     return ravendb.documents.DocumentCounters(self, entity_or_document_id)

    @staticmethod
    def raise_no_database() -> None:
        raise RuntimeError(
            "Cannot open a Session without specyfing a name of a database "
            "to operate on. Database name can be passed as an argument when Session is"
            " being opened or default database can be defined using 'DocumentStore.database' field"
        )

    @abstractmethod
    def _generate_id(self, entity: object) -> str:
        pass

    def _get_document_info(self, entity: object) -> DocumentInfo:
        document_info = self._documents_by_entity.get(entity)
        if document_info:
            return document_info

        found, value = self._generate_entity_id_on_client.try_get_id_from_instance(entity)
        if not found:
            raise ValueError(f"Could not find the document id for {entity}")
        self._assert_no_non_unique_instance(entity, value)
        raise ValueError(f"Document {value} doesn't exist in the session")

    def is_loaded(self, key: str) -> bool:
        return self.is_loaded_or_deleted(key)

    def is_loaded_or_deleted(self, key: str) -> bool:
        document_info = self._documents_by_id.get(key)
        return (
            document_info is not None
            and (document_info.document is not None or document_info.entity is not None)
            or self.is_deleted(key)
            or key in self._included_documents_by_id
        )

    def is_deleted(self, key: str) -> bool:
        return key in self._known_missing_ids

    def increment_requests_count(self) -> None:
        self._number_of_requests += 1
        if self._number_of_requests > self._max_number_of_requests_per_session:
            raise ValueError(
                f"The maximum number of requests {self._max_number_of_requests_per_session} allowed for this session"
                "has been reached. "
                "Raven limits the number of remote calls that a session is allowed to make as an early warning system. "
                "Sessions are expected to be short lived, and Raven provides facilities like load(String[] keys) "
                "to load multiple documents at once and batch saves (call SaveChanges() only once). "
                "You can increase the limit by setting DocumentConvention.MaxNumberOfRequestsPerSession or "
                "MaxNumberOfRequestsPerSession, but it is advisable that you'll look into reducing the number of remote"
                " calls first, since that will speed up your application significantly"
                "and result in a more responsive application."
            )

    def _assert_no_non_unique_instance(self, entity: object, key: str) -> None:
        if not key or key[-1] == "|" or key[-1] == self.conventions.identity_parts_separator:
            return

        info = self._documents_by_id.get(key)
        if info is None or info.entity is entity:
            return

        raise NonUniqueObjectException(f"Attempted to associate a different object with id '{key}'.")

    def track_entity_document_info(self, entity_type: Type[_T], document_found: DocumentInfo) -> _T:
        return self.track_entity(
            entity_type, document_found.key, document_found.document, document_found.metadata, self.no_tracking
        )

    def register_external_loaded_into_the_session(self, info: DocumentInfo) -> None:
        if self.no_tracking:
            return

        existing = self._documents_by_id.get_value(info.key)
        if existing is not None:
            if existing.entity == info.entity:
                return

            raise RuntimeError(f"The document {info.key} is already in the session with a different entity instance.")

        existing_entity = self._documents_by_entity.get(info.entity)
        if existing_entity is not None:
            if existing_entity.key.lower() == info.key.lower():
                return

            raise RuntimeError(
                f"Attempted to load an entity with id {info.key}, "
                f"but the entity instance already exists in the session with id: {existing_entity.key}"
            )

        self._documents_by_entity[info.entity] = info
        self._documents_by_id.add(info)
        self._included_documents_by_id.remove(info.key)

    def track_entity(
        self,
        entity_type: Type[_T],
        key: Optional[str],
        document: dict,
        metadata: dict,
        no_tracking: bool,
    ) -> _T:
        no_tracking = self.no_tracking or no_tracking
        if not key:
            return self.__deserialize_from_transformer(entity_type, None, document, False)

        doc_info = self._documents_by_id.get(key)
        if doc_info is not None:
            if doc_info.entity is None:
                doc_info.entity = self.entity_to_json.convert_to_entity(entity_type, key, document, not no_tracking)

            if not no_tracking:
                self._included_documents_by_id.pop(key, None)
                self._documents_by_entity[doc_info.entity] = doc_info

            return doc_info.entity

        doc_info = self._included_documents_by_id.get(key)
        if doc_info:
            if doc_info.entity is None:
                doc_info.entity = self.entity_to_json.convert_to_entity(entity_type, key, document, not no_tracking)

            if not no_tracking:
                self._included_documents_by_id.pop(key, None)
                self._documents_by_id[doc_info.key] = doc_info
                self._documents_by_entity[doc_info.entity] = doc_info

            return doc_info.entity

        entity = self.entity_to_json.convert_to_entity(entity_type, key, document, not no_tracking)

        change_vector = metadata.get(constants.Documents.Metadata.CHANGE_VECTOR)
        if change_vector is None:
            raise ValueError(f"Document {key} must have a Change Vector")

        if not no_tracking:
            new_document_info = DocumentInfo(
                key=key, document=document, metadata=metadata, entity=entity, change_vector=change_vector
            )
            self._documents_by_id[new_document_info.key] = new_document_info
            self._documents_by_entity[new_document_info.entity] = new_document_info

        return entity

    def delete(self, key_or_entity: Union[str, object], expected_change_vector: Optional[str] = None) -> None:
        if isinstance(key_or_entity, str):
            key = key_or_entity
            if key is None:
                raise ValueError("Id cannot be None")
            change_vector = None
            document_info = self._documents_by_id.get_value(key)
            if document_info is not None:
                new_obj = self.entity_to_json.convert_entity_to_json(document_info.entity, document_info)
                if document_info.entity is not None and self._entity_changed(new_obj, document_info, None):
                    raise InvalidOperationException(
                        "Can't delete changed entity using identifier. Use delete(entity) instead."
                    )

                if document_info.entity is not None:
                    self._documents_by_entity.pop(document_info.entity, None)
                self._documents_by_id.pop(key, None)
                change_vector = document_info.change_vector

            self._known_missing_ids.add(key)
            change_vector = change_vector if self._use_optimistic_concurrency else None
            if self._counters_by_doc_id:
                self._counters_by_doc_id.pop(key, None)
            self.defer(
                DeleteCommandData(
                    key,
                    expected_change_vector or change_vector,
                    expected_change_vector or (document_info.change_vector if document_info is not None else None),
                )
            )
            return

        entity = key_or_entity
        if entity is None:
            raise ValueError("Entity cannot be None")

        value = self._documents_by_entity.get(entity)
        if value is None:
            raise InvalidOperationException(
                f"{entity} is not associated with the session, cannot delete unknown entity instance."
            )

        self._deleted_entities.add(entity)
        self._included_documents_by_id.pop(value.key, None)
        if self._counters_by_doc_id:
            self._counters_by_doc_id.pop(value.key, None)
        self._known_missing_ids.add(value.key)

    def store(self, entity: object, key: Optional[str] = None, change_vector: Optional[str] = None) -> None:
        if all([entity, not key, not change_vector]):
            has_id = self._generate_entity_id_on_client.try_get_id_from_instance(entity)[0]
            checkmode = ConcurrencyCheckMode.FORCED if not has_id else ConcurrencyCheckMode.AUTO
        elif all([entity, key, not change_vector]):
            checkmode = ConcurrencyCheckMode.AUTO
        elif all([entity, key, change_vector]):
            checkmode = ConcurrencyCheckMode.DISABLED if change_vector is None else ConcurrencyCheckMode.FORCED
        else:
            raise ValueError("Unexpected arguments set. Pass entity / entity and key / entity,key and change_vector.")
        self.__store_internal(entity, change_vector, key, checkmode)

    def __store_internal(
        self, entity: object, change_vector: str, key: str, force_concurrency_check: ConcurrencyCheckMode
    ):
        if self.no_tracking:
            raise RuntimeError("Cannot legacy entity. Entity tracking is disabled in this session.")
        if entity is None:
            raise ValueError("Entity cannot be None")

        value = self._documents_by_entity.get(entity)
        if value is not None:
            value.change_vector = change_vector if change_vector else value.change_vector
            value.concurrency_check_mode = force_concurrency_check
            return

        if key is None:
            if self._generate_document_keys_on_store:
                key = self._generate_entity_id_on_client.generate_document_key_for_storage(entity)
            else:
                self._remember_entity_for_document_id_generation(entity)
        else:
            self._generate_entity_id_on_client.try_set_identity(entity, key)

        if IdTypeAndName.create(key, CommandType.CLIENT_ANY_COMMAND, None) in self._deferred_commands_map.keys():
            raise InvalidOperationException(
                f"Can't legacy document, there is a deferred command registered for this document in the session. "
                f"Document id:{key}"
            )

        if entity in self._deleted_entities:
            raise RuntimeError(f"Can't store object, it was already deleted in this session. Document id {key}")

        self._assert_no_non_unique_instance(entity, key)

        collection_name = (
            self.conventions.get_collection_name_for_dict(key)
            if entity.__class__ is dict and key and len(key.split("/")) > 1
            else self._request_executor.conventions.get_collection_name(entity)
        )

        metadata = {}
        if collection_name:
            metadata[constants.Documents.Metadata.COLLECTION] = collection_name
        python_type = self._request_executor.conventions.find_python_class_name(type(entity))
        if python_type:
            metadata[constants.Documents.Metadata.RAVEN_PYTHON_TYPE] = python_type
        if key:
            self._known_missing_ids.discard(key)
        self._store_entity_in_unit_of_work(key, entity, change_vector, metadata, force_concurrency_check)

    def _remember_entity_for_document_id_generation(self, entity: object) -> None:
        raise NotImplementedError(
            "You cannot set generate_document_ids_on_store to false without "
            "implementing remember_entity_for_document_id_generation"
        )
        pass

    def _store_entity_in_unit_of_work(
        self,
        key: str,
        entity: object,
        change_vector: str,
        metadata: dict,
        force_concurrency_check: ConcurrencyCheckMode,
    ) -> None:
        if key:
            self._known_missing_ids.discard(key)

        document_info = DocumentInfo(
            key=key,
            metadata=metadata,
            change_vector=change_vector,
            concurrency_check_mode=force_concurrency_check,
            entity=entity,
            new_document=True,
            document=None,
        )
        self._documents_by_entity[entity] = document_info
        if key is not None:
            self._documents_by_id[key] = document_info

    def prepare_for_save_changes(self) -> SaveChangesData:
        result = InMemoryDocumentSessionOperations.SaveChangesData(self)
        deferred_commands_count = len(self._deferred_commands)

        self.__prepare_for_entities_deletion(result, None)
        self.__prepare_for_entities_puts(result)
        self.__prepare_for_creating_revisions_from_ids(result)
        self.__prepare_compare_exchange_entities(result)

        if len(self._deferred_commands) > deferred_commands_count:
            # this allow on_before_store to call defer during the call to include
            # additional values during the same SaveChanges call

            result.deferred_commands.extend(self._deferred_commands)
            result.deferred_commands_map.update(self._deferred_commands_map)

        for deferred_command in result.deferred_commands:
            deferred_command.on_before_save_changes(self)

        return result

    def validate_cluster_transaction(self, result: SaveChangesData) -> None:
        if self.transaction_mode != TransactionMode.CLUSTER_WIDE:
            return

        if self._use_optimistic_concurrency:
            raise RuntimeError(
                f"useOptimisticConcurrency is not supported with TransactionMode set to {TransactionMode.CLUSTER_WIDE}"
            )

        for command_data in result.session_commands:
            if command_data.command_type == CommandType.PUT or CommandType.DELETE:
                if command_data.change_vector is not None:
                    raise ValueError(
                        f"Optimistic concurrency for {command_data.key} "
                        f"is not supported when using a cluster transaction"
                    )
            elif command_data.command_type == CommandType.COMPARE_EXCHANGE_DELETE or CommandType.COMPARE_EXCHANGE_PUT:
                pass
            else:
                raise ValueError(f"The command '{command_data.command_type}' is not supported in a cluster session.")

    def __prepare_compare_exchange_entities(self, result: SaveChangesData) -> None:
        if not self._has_cluster_session():
            return

        cluster_transaction_operations = self.cluster_transaction
        if cluster_transaction_operations.number_of_tracked_compare_exchange_values == 0:
            return

        if self.transaction_mode != TransactionMode.CLUSTER_WIDE:
            raise RuntimeError(
                "Performing cluster transaction operation require the TransactionMode to be set to CLUSTER_WIDE"
            )

        self.cluster_transaction.prepare_compare_exchange_entities(result)

    # --- CLUSTER ---
    @abstractmethod
    def _has_cluster_session(self) -> bool:
        pass

    @abstractmethod
    def _clear_cluster_session(self) -> None:
        pass

    @property
    @abstractmethod
    def cluster_transaction(self) -> ClusterTransactionOperationsBase:
        raise RuntimeError(f"{self.__class__.__name__} must override the cluster_session property method")

    def __prepare_for_creating_revisions_from_ids(self, result: SaveChangesData) -> None:
        for id_entry in self._ids_for_creating_forced_revisions:
            result.session_commands.append(ForceRevisionCommandData(id_entry))

        self._ids_for_creating_forced_revisions.clear()

    def __prepare_for_entities_deletion(
        self, result: Union[None, SaveChangesData], changes: Union[None, Dict[str, List[DocumentsChanges]]]
    ) -> None:
        for deleted_entity in self._deleted_entities:
            document_info = self._documents_by_entity.get(deleted_entity.entity)
            if document_info is None:
                continue
            if changes is not None:
                doc_changes = []
                change = DocumentsChanges("", "", DocumentsChanges.ChangeType.DOCUMENT_DELETED)
                doc_changes.append(change)
                changes[document_info.key] = doc_changes
            else:
                command = result.deferred_commands_map.get(
                    IdTypeAndName.create(document_info.key, CommandType.CLIENT_ANY_COMMAND, None)
                )
                if command:
                    self.__throw_invalid_deleted_document_with_deferred_command(command)

                change_vector = None
                document_info = self._documents_by_id.get(document_info.key)

                if document_info:
                    change_vector = document_info.change_vector

                    if document_info.entity is not None:
                        result.on_success.remove_document_by_entity(document_info.entity)
                        result.entities.append(document_info.entity)

                    result.on_success.remove_document_by_id(document_info.key)

                change_vector = change_vector if self._use_optimistic_concurrency else None
                self.before_delete_invoke(BeforeDeleteEventArgs(self, document_info.key, document_info.entity))
                result.session_commands.append(
                    DeleteCommandData(document_info.key, change_vector, document_info.change_vector)
                )

            if changes is None:
                result.on_success.clear_deleted_entities()

    def __prepare_for_entities_puts(self, result: SaveChangesData) -> None:
        should_ignore_entity_changes = self.conventions.should_ignore_entity_changes
        for entity in self._documents_by_entity:
            entity: DocumentsByEntityHolder.DocumentsByEntityEnumeratorResult
            if entity.value.ignore_changes:
                continue

            if should_ignore_entity_changes:
                if should_ignore_entity_changes.check(self, entity.value.entity, entity.value.key):
                    continue
            if self.is_deleted(entity.value.key):
                continue

            dirty_metadata = _update_metadata_modifications(entity.value.metadata_instance, entity.value.metadata)

            document = self.entity_to_json.convert_entity_to_json(entity.key, entity.value)

            if not self._entity_changed(document, entity.value, None) and not dirty_metadata:
                continue

            command = result.deferred_commands_map.get(
                IdTypeAndName.create(entity.value.key, CommandType.CLIENT_MODIFY_DOCUMENT_COMMAND, None),
                None,
            )
            if command:
                self.__throw_invalid_modified_document_with_deferred_command(command)

            if self._before_store and entity.execute_on_before_store:
                before_store_event_args = BeforeStoreEventArgs(self, entity.value.key, entity.key)
                self.before_store_invoke(before_store_event_args)

                if before_store_event_args.is_metadata_accessed:
                    _update_metadata_modifications(entity.value.metadata_instance, entity.value.metadata)

                if before_store_event_args.is_metadata_accessed or self._entity_changed(document, entity.value, None):
                    document = self.entity_to_json.convert_entity_to_json(entity.key, entity.value)

            result.entities.append(entity.key)

            if entity.value.key is not None:
                result.on_success.remove_document_by_id(entity.value.key)

            result.on_success.update_entity_document_info(entity.value, document)

            change_vector = (
                (entity.value.change_vector if entity.value.change_vector else "")
                if self._use_optimistic_concurrency
                else (
                    entity.value.change_vector
                    if entity.value.concurrency_check_mode == ConcurrencyCheckMode.FORCED
                    else None
                )
            )

            force_revision_creation_strategy = ForceRevisionStrategy.NONE
            if entity.value.key is not None:
                creation_strategy = self._ids_for_creating_forced_revisions.get(entity.value.key)
                if creation_strategy is not None:
                    self._ids_for_creating_forced_revisions.pop(entity.value.key, None)
                    force_revision_creation_strategy = creation_strategy

            result.session_commands.append(
                PutCommandDataWithJson(entity.value.key, change_vector, document, force_revision_creation_strategy)
            )

    @staticmethod
    def __throw_invalid_modified_document_with_deferred_command(result_command: CommandData) -> None:
        raise RuntimeError(
            f"Cannot perform save because document {result_command.key} has been modified by "
            f"the session and is also taking part in deferred {result_command.command_type} command"
        )

    @staticmethod
    def __throw_invalid_deleted_document_with_deferred_command(result_command: CommandData) -> None:
        raise RuntimeError(
            f"Cannot perform save because document {result_command.key} has been deleted by "
            f"the session and is also taking part in deferred {result_command.command_type} command"
        )

    @staticmethod
    def __throw_no_database() -> None:
        raise RuntimeError(
            "Cannot open a Session without specyifing a name of a database to operate on. "
            "Database name can be passed as an argument when Session is being opened "
            "or default database can be defined using legacy.database property"
        )

    def _entity_changed(
        self, new_obj: dict, document_info: DocumentInfo, changes: Union[None, Dict[str, List[DocumentsChanges]]]
    ) -> bool:
        return JsonOperation.entity_changed(new_obj, document_info, changes)

    def has_changes(self) -> bool:
        for entity in self._documents_by_entity:
            entity: DocumentsByEntityHolder.DocumentsByEntityEnumeratorResult
            document = self.entity_to_json.convert_entity_to_json(entity.key, entity.value)
            if self._entity_changed(document, entity.value, None):
                return True

        return not len(self._deleted_entities) == 0 or not len(self._deferred_commands) == 0

    def _what_changed(self) -> Dict[str, List[DocumentsChanges]]:
        changes = {}

        self.__get_all_entities_changes(changes)
        self.__prepare_for_entities_deletion(None, changes)

        return changes

    def __get_all_entities_changes(self, changes: Dict[str, List[DocumentsChanges]]) -> None:
        for key, value in self._documents_by_id.items():
            _update_metadata_modifications(value.metadata_instance, value.metadata)
            new_obj = self.entity_to_json.convert_entity_to_json(value.entity, value)
            self._entity_changed(new_obj, value, changes)

    def ignore_changes_for(self, entity: object) -> None:
        self._get_document_info(entity).ignore_changes = True

    def defer(self, *commands: CommandData) -> None:
        self._deferred_commands.extend(commands)
        for command in commands:
            self._defer_internal(command)

    def _defer_internal(self, command: CommandData) -> None:
        if command.command_type is CommandType.BATCH_PATCH:
            command: BatchPatchCommandData
            for kvp in command.ids:
                self.__add_command(command, kvp.key, CommandType.PATCH, command.name)
            return
        self.__add_command(command, command.key, command.command_type, command.name)

    def __add_command(self, command: CommandData, key: str, command_type: CommandType, command_name: str) -> None:
        self._deferred_commands_map.update({IdTypeAndName.create(key, command_type, command_name): command})
        self._deferred_commands_map.update({IdTypeAndName.create(key, CommandType.CLIENT_ANY_COMMAND, None): command})

        if not any(
            [
                command.command_type == CommandType.ATTACHMENT_DELETE,
                command.command_type == CommandType.TIME_SERIES_COPY,
                command.command_type == CommandType.ATTACHMENT_MOVE,
                command.command_type == CommandType.ATTACHMENT_COPY,
                command.command_type == CommandType.ATTACHMENT_PUT,
                command.command_type == CommandType.TIME_SERIES,
                command.command_type == CommandType.COUNTERS,
            ]
        ):
            self._deferred_commands_map.update(
                {IdTypeAndName.create(key, CommandType.CLIENT_MODIFY_DOCUMENT_COMMAND, None): command}
            )

    def __close(self, is_disposing: bool) -> None:
        if self._is_disposed:
            return

        self.session_closing_invoke(SessionClosingEventArgs(self))
        self._is_disposed = True

    def close(self) -> None:
        self.__close(True)

    def register_missing(self, *keys: str) -> None:
        if self.no_tracking:
            return
        self._known_missing_ids.update(keys)

    def register_includes(self, includes: dict):
        if self.no_tracking:
            return

        if not includes:
            return

        for key, value in includes.items():
            if value is None:
                continue
            object_node = json.loads(json.dumps(value))
            new_document_info = DocumentInfo.get_new_document_info(object_node)
            if JsonExtensions.try_get_conflict(new_document_info.metadata):
                continue
            self._included_documents_by_id[new_document_info.key] = new_document_info

    def register_missing_includes(self, results, includes: dict, include_paths: List[str]):
        if self.no_tracking:
            return

        if not include_paths:
            return

        def __include(key):
            if not key:
                return
            if self.is_loaded(key):
                return

            document = includes.get(key)
            if document is not None:
                metadata = document.get(constants.Documents.Metadata.KEY)

                if JsonExtensions.try_get_conflict(metadata):
                    return

            self.register_missing()

        for result in results:
            for include in include_paths:
                if constants.Documents.Indexing.Fields.DOCUMENT_ID_FIELD_NAME == include:
                    continue
                IncludesUtil.include(result, include, __include)

    def register_counters_ids_list(
        self, result_counters: dict, keys: List[str], counters_to_include: List[str], got_all: bool
    ) -> None:
        if self.no_tracking:
            return

        if not result_counters:
            if got_all:
                for key in keys:
                    self.__set_got_all_counters_for_document(key)

                return
        else:
            self._register_counters_internal(result_counters, None, False, got_all)

        self.__register_missing_counters_for_keys(keys, counters_to_include)

    def register_counters(
        self,
        result_counters: dict,
        counters_to_include: Dict[str, List[str]],
    ) -> None:
        if self.no_tracking:
            return

        if not result_counters:
            self.__set_got_all_in_cache_if_needed(counters_to_include)
        else:
            self._register_counters_internal(result_counters, counters_to_include, True, False)

        self.__register_missing_counters(counters_to_include)

    def _register_counters_internal(
        self,
        result_counters: dict,
        counters_to_include: Optional[Dict[str, List[str]]],
        from_query_result: bool,
        got_all: bool,
    ):
        for key, result_counters in result_counters.items():
            if not result_counters:
                continue
            counters = []
            if from_query_result:
                counters = counters_to_include.get(key, None)
                got_all = True if counters is not None and len(counters) == 0 else False

            if len(result_counters) == 0 and not got_all:
                cache = self._counters_by_doc_id.get(key, None)
                if not cache:
                    continue
                for counter in counters:
                    del cache[1][counter]
                self._counters_by_doc_id[key] = cache
                continue
            self.__register_counters_for_document(key, got_all, result_counters, counters_to_include)

    def __register_counters_for_document(
        self, key: str, got_all: bool, result_counters: List[Dict], counters_to_include: Dict[str, List[str]]
    ):
        cache = self._counters_by_doc_id.get(key)
        if not cache:
            cache = [got_all, CaseInsensitiveDict()]

        deleted_counters = (
            set()
            if len(cache[1]) == 0
            else (set(cache[1].keys()) if len(counters_to_include.get(key)) == 0 else set(counters_to_include.get(key)))
        )

        for counter in result_counters:
            if counter:
                counter_name = counter["CounterName"]
                total_value = counter["TotalValue"]
            else:
                counter_name = total_value = None
            if counter_name and total_value:
                cache[1][counter_name] = total_value
                deleted_counters.discard(counter_name)

        if deleted_counters:
            for name in deleted_counters:
                del cache[1][name]

        cache[0] = got_all
        self._counters_by_doc_id[key] = cache

    def __set_got_all_in_cache_if_needed(self, counters_to_include: Dict[str, List[str]]):
        if not counters_to_include:
            return
        for key, value in counters_to_include:
            if len(value) > 0:
                continue
            self.__set_got_all_counters_for_document(key)

    def __set_got_all_counters_for_document(self, key: str):
        cache = self._counters_by_doc_id.get(key, None)
        if not cache:
            cache = [False, CaseInsensitiveDict()]
        cache[0] = True
        self._counters_by_doc_id[key] = cache

    def __register_missing_counters(self, counters_to_include: Dict[str, List[str]]):
        if not counters_to_include:
            return

        for key, value in counters_to_include.items():
            cache = self._counters_by_doc_id.get(key, None)
            if cache is None:
                cache = [False, CaseInsensitiveDict()]
                self._counters_by_doc_id[key] = cache

            for counter in value:
                if counter in cache[1]:
                    continue
                cache[1][counter] = None

    def __register_missing_counters_for_keys(self, keys: List[str], counters_to_include: List[str]):
        if not counters_to_include:
            return

        for counter in counters_to_include:
            for key in keys:
                cache = self._counters_by_doc_id.get(key, None)
                if cache is None:
                    cache = [False, CaseInsensitiveDict()]
                    self._counters_by_doc_id[key] = cache

                if counter in cache[1]:
                    continue

                cache[1][counter] = None

    def register_time_series(self, result_time_series: Dict[str, Dict[str, List[Dict[str, Any]]]]):
        if self.no_tracking or not result_time_series:
            return

        for doc_id, time_series in result_time_series.items():
            if not time_series:
                continue

            # compute if absent (cache the result)
            session_cache = self._time_series_by_doc_id
            cached_value = None
            if doc_id in session_cache:
                cached_value = session_cache[doc_id]
            if cached_value is None:
                cached_value = CaseInsensitiveDict()
                session_cache[doc_id] = cached_value

            for time_series_name, time_series_ranges in time_series.items():
                if not time_series_ranges:
                    continue
                name = time_series_name
                for time_series_range in time_series_ranges:
                    time_series_range_result = TimeSeriesRangeResult.from_json(time_series_range)
                    self.__add_to_cache(cached_value, time_series_range_result, name)

    @staticmethod
    def __add_to_cache(cache: Dict[str, List[TimeSeriesRangeResult]], new_range: TimeSeriesRangeResult, name: str):
        local_ranges = cache.get(name)
        if not local_ranges:
            # No local ranges in cache for this series
            cache[name] = [new_range]
            return

        if TSRangeHelper.left(local_ranges[0].from_date) > TSRangeHelper.right(
            new_range.to_date
        ) or TSRangeHelper.right(local_ranges[-1].to_date) < TSRangeHelper.left(new_range.from_date):
            # the entire range [from, to] is out of cache bounds
            index = (
                0
                if TSRangeHelper.left(local_ranges[0].from_date) > TSRangeHelper.right(new_range.to_date)
                else len(local_ranges)
            )
            local_ranges.insert(index, new_range)
            return

        to_range_index = int()
        from_range_index = -1
        range_already_in_cache = False
        broke_out_of_loop = False

        for to_range_index in range(len(local_ranges)):
            if TSRangeHelper.left(local_ranges[to_range_index].from_date) <= TSRangeHelper.left(new_range.from_date):
                if TSRangeHelper.right(local_ranges[to_range_index].to_date) >= TSRangeHelper.right(new_range.to_date):
                    range_already_in_cache = True
                    broke_out_of_loop = True
                    break

                from_range_index = to_range_index
                continue

            if TSRangeHelper.right(local_ranges[to_range_index].to_date) >= TSRangeHelper.right(new_range.to_date):
                broke_out_of_loop = True
                break

        if not broke_out_of_loop:
            to_range_index += 1  # in Java, the last increment happens before the condition check

        if range_already_in_cache:
            InMemoryDocumentSessionOperations.__update_existing_range(local_ranges[to_range_index], new_range)
            return

        merged_values = InMemoryDocumentSessionOperations.__merge_ranges(
            from_range_index, to_range_index, local_ranges, new_range
        )
        InMemoryDocumentSessionOperations.add_to_cache(
            name,
            new_range.from_date,
            new_range.to_date,
            from_range_index,
            to_range_index,
            local_ranges,
            cache,
            merged_values,
        )

    @staticmethod
    def add_to_cache(
        time_series: str,
        from_date: datetime.datetime,
        to_date: datetime.datetime,
        from_range_index: int,
        to_range_index: int,
        ranges: List[TimeSeriesRangeResult],
        cache: Dict[str, List[TimeSeriesRangeResult]],
        values: List[TimeSeriesEntry],
    ):
        if from_range_index == -1:
            # didn't find a 'from_range' => all ranges in cache start after 'from'

            if to_range_index == len(ranges):
                # the requested range [from, to] contains all the ranges that are in cache

                # e.g. if cache is : [[2,3], [4,5], [7,10]]
                # and the requested range is : [1,15]
                # after this action cache will be : [[1, 15]]

                time_series_range_result = TimeSeriesRangeResult(from_date, to_date, values)
                result = [time_series_range_result]
                cache[time_series] = result
                return

            if TSRangeHelper.left(ranges[to_range_index].from_date) > TSRangeHelper.right(to_date):
                # requested range ends before 'toRange' starts
                # remove all ranges that come before 'toRange' from cache
                # add the new range at the beginning of the list

                # e.g. if cache is : [[2,3], [4,5], [7,10]]
                # and the requested range is : [1,6]
                # after this action cache will be : [[1,6], [7,10]]

                for i in range(0, to_range_index):
                    del ranges[0]

                time_series_range_result = TimeSeriesRangeResult(from_date, to_date, values)
                ranges.insert(0, time_series_range_result)
                return

            # the requested range ends inside 'toRange'
            # merge the result from server into 'toRange'
            # remove all ranges that come before 'toRange' from cache

            # e.g. if cache is : [[2,3], [4,5], [7,10]]
            # and the requested range is : [1,8]
            # after this action cache will be : [[1,10]]

            ranges[to_range_index].from_date = from_date
            ranges[to_range_index].entries = values

            for i in range(0, to_range_index):
                del ranges[0]
            return

        # found a 'from_range'

        if to_range_index == len(ranges):
            # didn't find a 'to_range' => all the ranges in cache end before 'to'

            if TSRangeHelper.left(from_date) > TSRangeHelper.right(ranges[from_range_index].to_date):
                # requested range starts after 'fromRange' ends,
                # so it needs to be placed right after it
                # remove all the ranges that come after 'fromRange' from cache
                # add the merged values as a new range at the end of the list

                # e.g. if cache is : [[2,3], [5,6], [7,10]]
                # and the requested range is : [4,12]
                # then 'fromRange' is : [2,3]
                # after this action cache will be : [[2,3], [4,12]]

                for i in range(from_range_index + 1, len(ranges)):
                    del ranges[from_range_index + 1]

                time_series_range_result = TimeSeriesRangeResult(from_date, to_date, values)
                ranges.append(time_series_range_result)
                return

            # the requested range starts inside 'fromRange'
            # merge result into 'fromRange'
            # remove all the ranges from cache that come after 'fromRange'

            # e.g. if cache is : [[2,3], [4,6], [7,10]]
            # and the requested range is : [5,12]
            # then 'fromRange' is [4,6]
            # after this action cache will be : [[2,3], [4,12]]

            ranges[from_range_index].to_date = to_date
            ranges[from_range_index].entries = values

            for i in range(from_range_index + 1, len(ranges)):
                del ranges[from_range_index + 1]
            return

        # found both 'from_range' and 'to_range'
        # the requested range is inside cache bounds

        if TSRangeHelper.left(from_date) > TSRangeHelper.right(ranges[from_range_index].to_date):
            # requested range starts after 'from_range' ends

            if TSRangeHelper.left(ranges[to_range_index].from_date) > TSRangeHelper.right(to_date):
                # requested range ends before 'toRange' starts

                # remove all ranges in between 'fromRange' and 'toRange'
                # place new range in between 'fromRange' and 'toRange'

                # e.g. if cache is : [[2,3], [5,6], [7,8], [10,12]]
                # and the requested range is : [4,9]
                # then 'fromRange' is [2,3] and 'toRange' is [10,12]
                # after this action cache will be : [[2,3], [4,9], [10,12]]

                for i in range(from_range_index + 1, to_range_index):
                    del ranges[from_range_index + 1]

                time_series_range_result = TimeSeriesRangeResult(from_date, to_date, values)
                ranges.insert(from_range_index + 1, time_series_range_result)
                return

            # requested range ends inside 'toRange'

            # merge the new range into 'toRange'
            # remove all ranges in between 'fromRange' and 'toRange'

            # e.g. if cache is : [[2,3], [5,6], [7,10]]
            # and the requested range is : [4,9]
            # then 'fromRange' is [2,3] and 'toRange' is [7,10]
            # after this action cache will be : [[2,3], [4,10]]

            for i in range(from_range_index + 1, to_range_index):
                del ranges[from_range_index + 1]
            ranges[to_range_index].from_date = from_date
            ranges[to_range_index].entries = values

            return

        # the requested range starts inside 'from_range'

        if TSRangeHelper.left(ranges[to_range_index].from_date) > TSRangeHelper.right(to_date):
            # requested range ends before 'toRange' starts

            # remove all ranges in between 'fromRange' and 'toRange'
            # merge new range into 'fromRange'

            # e.g. if cache is : [[2,4], [5,6], [8,10]]
            # and the requested range is : [3,7]
            # then 'fromRange' is [2,4] and 'toRange' is [8,10]
            # after this action cache will be : [[2,7], [8,10]]

            ranges[from_range_index].to_date = to_date
            ranges[from_range_index].entries = values
            for i in range(from_range_index + 1, to_range_index):
                del ranges[from_range_index + 1]
            return

        # the requested range starts inside 'fromRange'
        # and ends inside 'toRange'

        # merge all ranges in between 'fromRange' and 'toRange'
        # into a single range [fromRange.From, toRange.To]

        # e.g. if cache is : [[2,4], [5,6], [8,10]]
        # and the requested range is : [3,9]
        # then 'fromRange' is [2,4] and 'toRange' is [8,10]
        # after this action cache will be : [[2,10]]

        ranges[from_range_index].to_date = ranges[to_range_index].to_date
        ranges[from_range_index].entries = values
        for i in range(from_range_index + 1, to_range_index + 1):
            del ranges[from_range_index + 1]

    @staticmethod
    def __merge_ranges(
        from_range_index: int,
        to_range_index: int,
        local_ranges: List[TimeSeriesRangeResult],
        new_range: TimeSeriesRangeResult,
    ) -> List[TimeSeriesEntry]:
        merged_values = []
        if from_range_index != -1 and local_ranges[from_range_index].to_date >= new_range.from_date:
            for val in local_ranges[from_range_index].entries:
                if val.timestamp >= new_range.from_date:
                    break
                merged_values.append(val)

        merged_values.extend(new_range.entries)

        if to_range_index < len(local_ranges) and TSRangeHelper.left(
            local_ranges[to_range_index].from_date
        ) <= TSRangeHelper.right(new_range.to_date):
            for val in local_ranges[to_range_index].entries:
                if val.timestamp <= new_range.to_date:
                    continue
                merged_values.append(val)
        return merged_values

    @staticmethod
    def __update_existing_range(local_range: TimeSeriesRangeResult, new_range: TimeSeriesRangeResult) -> None:
        new_values = []
        for index in range(len(local_range.entries)):
            if local_range.entries[index].timestamp >= new_range.from_date:
                break

            new_values.append(local_range.entries[index])

        new_values.extend(new_range.entries)

        for j in range(len(local_range.entries)):
            if local_range.entries[j].timestamp <= new_range.to_date:
                continue
            new_values.append(local_range.entries[j])
        local_range.entries = new_values

    def hash_code(self) -> int:
        return self._hash

    def __deserialize_from_transformer(
        self, object_type: Type[_T], key: Union[None, str], document: dict, track_entity: bool
    ) -> _T:
        return self.entity_to_json.convert_to_entity(object_type, key, document, track_entity)

    def check_if_all_change_vectors_are_already_included(self, change_vectors: List[str]) -> bool:
        if self.include_revisions_by_change_vector is None:
            return False

        for cv in change_vectors:
            if cv not in self.include_revisions_by_change_vector:
                return False

        return True

    def check_if_revisions_by_date_time_before_already_included(self, id_: str, date_time: datetime) -> bool:
        if self.include_revisions_by_date_time_before is None:
            return False

        dictionary_date_time_to_document = self.include_revisions_by_date_time_before.get(id_)
        if dictionary_date_time_to_document is not None:
            return date_time in dictionary_date_time_to_document

        return False

    def check_if_id_already_included(self, ids: List[str], includes: Union[List[List], List[str]]) -> bool:
        if includes and not isinstance(includes[0], str):
            return self.check_if_id_already_included(ids, [arr[1] for arr in includes])
        includes: List[str]
        for key in ids:
            if key in self._known_missing_ids:
                continue

            document_info = self._documents_by_id.get(key)
            if document_info is None:
                document_info = self._included_documents_by_id.get(key)
                if document_info is None:
                    return False

            if document_info.entity is None and document_info.document is None:
                return False

            if includes is None:
                continue

            for include in includes:
                has_all = [True]
                # todo: make sure it'll work - we create has_all and praise for changes on the reference has_all[0]
                IncludesUtil.include(document_info.document, include, lambda s: has_all[0] and self.is_loaded(s))
                if not has_all[0]:
                    return False
        return True

    def _refresh_internal(self, entity: object, cmd: RavenCommand, document_info: DocumentInfo):
        if not isinstance(cmd.result, GetDocumentsResult):
            raise TypeError(f"Unexpected RavenCommand.result type '{type(cmd.result)}' - expected 'GetDocumentResult'")
        try:
            document = cmd.result.results[0]
        except IndexError:
            raise ValueError(f"Document {document_info.key} no longer exists and was probably deleted")

        value = document.get(constants.Documents.Metadata.KEY)
        document_info.metadata = value

        if document_info.metadata is not None:
            change_vector = value.get(constants.Documents.Metadata.CHANGE_VECTOR)
            document_info.change_vector = change_vector

        if document_info.entity is not None and not self.no_tracking:
            self.entity_to_json.remove_from_missing(document_info.entity)

        document_info.entity = self.entity_to_json.convert_to_entity(
            type(entity), document_info.key, document, not self.no_tracking
        )
        document_info.document = document

        try:
            entity = deepcopy(document_info.entity)
        except Error as e:
            raise RuntimeError(f"Unable to refresh entity: {e.args[0]}", e)

        document_info_by_id = self._documents_by_id.get(document_info.key)
        if document_info_by_id is not None:
            document_info_by_id.entity = entity
        return entity

    def _get_operation_result(self, object_type: Type[_T], result: _T) -> _T:
        if result is None:
            return Utils.get_default_value(object_type)

        if issubclass(type(result), object_type) or type(result) == object_type:
            return result
            # todo: cast result on object_type
        if isinstance(result, dict):
            if not result:
                return None
            return result.values().__iter__().__next__()
            # todo: cast result on object_type
        raise TypeError(f"Unable to cast {result.__class__.__name__} to {object_type.__name__}")

    # todo: implement method below
    def update_session_after_save_changes(self, result: BatchCommandResult):
        returned_transaction_index = result.transaction_index

    def _process_query_parameters(
        self, object_type: type, index_name: str, collection_name: str, conventions: DocumentConventions
    ) -> (str, str):
        is_index = index_name and not index_name.isspace()
        is_collection = collection_name and not collection_name.isspace()

        if is_index and is_collection:
            raise ValueError(
                "Parameters index_name and collection_name are mutually exclusive. Please specify only one."
            )

        if not is_collection and not is_index:
            collection_name = conventions.get_collection_name(object_type)
            collection_name = (
                collection_name if collection_name else constants.Documents.Metadata.ALL_DOCUMENTS_COLLECTION
            )

        return index_name, collection_name

    class ReplicationWaitOptsBuilder:
        def __init__(self, session: InMemoryDocumentSessionOperations):
            self.__session = session

        def get_options(self) -> BatchOptions:
            if self.__session._save_changes_options is None:
                self.__session._save_changes_options = BatchOptions()

            if self.__session._save_changes_options.replication_options is None:
                self.__session._save_changes_options.replication_options = ReplicationBatchOptions()

            return self.__session._save_changes_options

        def with_timeout(
            self, timeout: datetime.timedelta
        ) -> InMemoryDocumentSessionOperations.ReplicationWaitOptsBuilder:
            self.get_options().replication_options.wait_for_indexes_timeout = timeout
            return self

        def throw_on_timeout(self, should_throw: bool) -> InMemoryDocumentSessionOperations.ReplicationWaitOptsBuilder:
            self.get_options().replication_options.throw_on_timeout_in_wait_for_replicas = should_throw
            return self

        def number_of_replicas(self, replicas: int) -> InMemoryDocumentSessionOperations.ReplicationWaitOptsBuilder:
            self.get_options().replication_options.number_of_replicas_to_wait_for = replicas
            return self

        def majority(self, wait_for_majority: bool) -> InMemoryDocumentSessionOperations.ReplicationWaitOptsBuilder:
            self.get_options().replication_options.majority = wait_for_majority
            return self

    class IndexesWaitOptsBuilder:
        def __init__(self, session: InMemoryDocumentSessionOperations):
            self.__session = session

        def get_options(self) -> BatchOptions:
            if self.__session._save_changes_options is None:
                self.__session._save_changes_options = BatchOptions()

            if self.__session._save_changes_options.index_options is None:
                self.__session._save_changes_options.index_options = IndexBatchOptions()

            return self.__session._save_changes_options

        def with_timeout(self, timeout: datetime.timedelta) -> InMemoryDocumentSessionOperations.IndexesWaitOptsBuilder:
            self.get_options().index_options.wait_for_indexes_timeout = timeout
            return self

        def throw_on_timeout(self, should_throw: bool) -> InMemoryDocumentSessionOperations.IndexesWaitOptsBuilder:
            self.get_options().index_options.throw_on_timeout_in_wait_for_replicas = should_throw
            return self

        def wait_for_indexes(self, *indexes: str) -> InMemoryDocumentSessionOperations.IndexesWaitOptsBuilder:
            self.get_options().index_options.wait_for_indexes = indexes
            return self

    class SaveChangesData:
        def __init__(self, session: InMemoryDocumentSessionOperations):
            self.deferred_commands: List[CommandData] = session._deferred_commands
            self.deferred_commands_map: Dict[IdTypeAndName, CommandData] = session._deferred_commands_map
            self.session_commands: List[CommandData] = []
            self.entities: List = []
            self.options = session._save_changes_options
            self.on_success = InMemoryDocumentSessionOperations.SaveChangesData.ActionsToRunOnSuccess(session)

        class ActionsToRunOnSuccess:
            def __init__(self, session: InMemoryDocumentSessionOperations):
                self.__session = session
                self.__documents_by_id_to_remove: List[str] = []
                self.__documents_by_entity_to_remove: List = []
                self.__document_infos_to_update: List[Tuple[DocumentInfo, dict]] = []
                self.__clear_deleted_entities: bool = False

            def remove_document_by_id(self, key: str):
                self.__documents_by_id_to_remove.append(key)

            def remove_document_by_entity(self, entity):
                self.__documents_by_entity_to_remove.append(entity)

            def update_entity_document_info(self, document_info: DocumentInfo, document: dict):
                self.__document_infos_to_update.append((document_info, document))

            def clear_session_state_after_successful_save_changes(self):
                for key in self.__documents_by_id_to_remove:
                    if key in self.__session.documents_by_id:
                        self.__session._documents_by_id.pop(key)
                for key in self.__documents_by_entity_to_remove:
                    self.__session._documents_by_entity.pop(key)

                for document_info_dict_tuple in self.__document_infos_to_update:
                    info: DocumentInfo = document_info_dict_tuple[0]
                    document: dict = document_info_dict_tuple[1]
                    info.new_document = False
                    info.document = document

                if self.__clear_deleted_entities:
                    self.__session._deleted_entities.clear()

                self.__session._deferred_commands.clear()
                self.__session._deferred_commands_map.clear()

            def clear_deleted_entities(self) -> None:
                self.__clear_deleted_entities = True



from typing import Optional, Generic, TypeVar, List

from ravendb.documents.operations.connection_strings import ConnectionString
import ravendb.serverwide

_T = TypeVar("_T")


class RavenConnectionString(ConnectionString):
    def __init__(self, name: str, database: Optional[str] = None, topology_discovery_urls: Optional[List[str]] = None):
        super().__init__(name)
        self.database = database
        self.topology_discovery_urls = topology_discovery_urls

    @property
    def get_type(self):
        return ravendb.serverwide.ConnectionStringType.RAVEN


# todo: implement
class EtlConfiguration(ConnectionString, Generic[_T]):
    pass


class RavenEtlConfiguration(EtlConfiguration[RavenConnectionString]):
    pass



import sys

from ravendb.documents.indexes.vector.embedding import VectorEmbeddingType

int_max = 0x7FFFFFFF
int_min = -int_max - 1
min_normal = sys.float_info.min
json_serialize_method_name = "to_json"
nan_value = float("nan")


class Documents:
    class Metadata:
        COLLECTION = "@collection"
        CONFLICT = "@conflict"
        PROJECTION = "@projection"
        METADATA = "@metadata"
        KEY = "@metadata"
        ID = "@id"
        FLAGS = "@flags"
        ATTACHMENTS = "@attachments"
        INDEX_SCORE = "@index-score"
        LAST_MODIFIED = "@last-modified"
        CHANGE_VECTOR = "@change-vector"
        EXPIRES = "@expires"
        REFRESH = "@refresh"
        ALL_DOCUMENTS_COLLECTION = "@all_docs"
        EMPTY_COLLECTION = "@empty"
        NESTED_OBJECT_TYPES = "@nested-object-types"
        COUNTERS = "@counters"
        TIME_SERIES = "@timeseries"
        REVISION_COUNTERS = "@counters-snapshot"
        REVISION_TIME_SERIES = "@timeseries-snapshot"
        RAVEN_PYTHON_TYPE = "Raven-Python-Type"

    class Indexing:
        SIDE_BY_SIDE_INDEX_NAME_PREFIX = "ReplacementOf/"

        class Fields:
            DOCUMENT_ID_FIELD_NAME = "id()"
            SOURCE_DOCUMENT_ID_FIELD_NAME = "sourceDocId()"
            REDUCE_KEY_HASH_FIELD_NAME = "hash(key())"
            REDUCE_KEY_KEY_VALUE_FIELD_NAME = "key()"
            VALUE_FIELD_NAME = "value()"
            ALL_FIELDS = "__all_fields"
            SPATIAL_SHAPE_FIELD_NAME = "spatial(shape)"

            class JavaScript:
                VECTOR_PROPERTY_NAME = "$vector"

        class Spatial:
            DEFAULT_DISTANCE_ERROR_PCT = 0.025


class CompareExchange:
    OBJECT_FIELD_NAME = "Object"


class Counters:
    ALL = "@all_counters"


class Headers:
    REQUEST_TIME = "Raven-Request-Time"
    REFRESH_TOPOLOGY = "Refresh-Topology"
    TOPOLOGY_ETAG = "Topology-Etag"
    LAST_KNOWN_CLUSTER_TRANSACTION_INDEX = "Known-Raft-Index"
    CLIENT_CONFIGURATION_ETAG = "Client-Configuration-Etag"
    REFRESH_CLIENT_CONFIGURATION = "Refresh-Client-Configuration"
    CLIENT_VERSION = "Raven-Client-Version"
    SERVER_VERSION = "Raven-Server-Version"
    ETAG = "ETag"
    IF_NONE_MATCH = "If-None-Match"
    TRANSFER_ENCODING = "Transfer-Encoding"
    CONTENT_ENCODING = "Content-Encoding"
    CONTENT_LENGTH = "Content-Length"


class TimeSeries:
    SELECT_FIELD_NAME = "timeseries"
    QUERY_FUNCTION = "__timeSeriesQueryFunction"

    ALL = "@all_timeseries"


class VectorSearch:
    EMBEDDING_PREFIX = "embedding."
    EMBEDDING_TEXT = EMBEDDING_PREFIX + "text"
    EMBEDDING_TEXT_INT_8 = EMBEDDING_PREFIX + "text_i8"
    EMBEDDING_TEXT_INT_1 = EMBEDDING_PREFIX + "text_i1"
    EMBEDDING_SINGLE = EMBEDDING_PREFIX + "f32"
    EMBEDDING_SINGLE_INT8 = EMBEDDING_PREFIX + "f32_i8"
    EMBEDDING_SINGLE_INT1 = EMBEDDING_PREFIX + "f32_i1"
    EMBEDDING_INT8 = EMBEDDING_PREFIX + "i8"
    EMBEDDING_INT1 = EMBEDDING_PREFIX + "i1"

    @staticmethod
    def configuration_to_method_name(source: VectorEmbeddingType, dest: VectorEmbeddingType):
        mapping = {
            (VectorEmbeddingType.SINGLE, VectorEmbeddingType.SINGLE): "",
            (VectorEmbeddingType.SINGLE, VectorEmbeddingType.INT8): VectorSearch.EMBEDDING_SINGLE_INT8,
            (VectorEmbeddingType.SINGLE, VectorEmbeddingType.BINARY): VectorSearch.EMBEDDING_SINGLE_INT1,
            (VectorEmbeddingType.TEXT, VectorEmbeddingType.SINGLE): VectorSearch.EMBEDDING_TEXT,
            (VectorEmbeddingType.TEXT, VectorEmbeddingType.INT8): VectorSearch.EMBEDDING_TEXT_INT_8,
            (VectorEmbeddingType.TEXT, VectorEmbeddingType.BINARY): VectorSearch.EMBEDDING_TEXT_INT_1,
            (VectorEmbeddingType.INT8, VectorEmbeddingType.INT8): VectorSearch.EMBEDDING_INT8,
            (VectorEmbeddingType.BINARY, VectorEmbeddingType.BINARY): VectorSearch.EMBEDDING_INT1,
        }
        if (source, dest) not in mapping:
            raise ValueError(
                f"Invalid embedding configuration. SourceEmbedding: {source.value}, DestinationEmbedding: {dest.value}"
            )
        return mapping[(source, dest)]

    DEFAULT_EMBEDDING_TYPE = VectorEmbeddingType.SINGLE
    DEFAULT_IS_EXACT = False



from __future__ import annotations

import enum
import os
from typing import List, Union, Optional, Tuple

from ravendb.documents.indexes.vector.embedding import VectorEmbeddingType
from ravendb.documents.session.time_series import (
    TimeSeriesRange,
    TimeSeriesTimeRange,
    TimeSeriesRangeType,
    AbstractTimeSeriesRange,
    TimeSeriesCountRange,
)
from ravendb.primitives import constants
from ravendb.documents.session.misc import OrderingType
from ravendb.documents.indexes.spatial.configuration import SpatialUnits
from ravendb.documents.queries.group_by import GroupByMethod
from ravendb.documents.queries.misc import SearchOperator
from ravendb.documents.queries.query import QueryOperator
from ravendb.documents.session.tokens.misc import WhereOperator
from ravendb.documents.session.tokens.query_tokens.query_token import QueryToken
from ravendb.documents.session.utils.document_query import DocumentQueryHelper
from ravendb.primitives.constants import VectorSearch
from ravendb.tools.utils import Utils


class CompareExchangeValueIncludesToken(QueryToken):
    def __init__(self, path: str):
        if not path:
            raise ValueError("Path cannot be None")
        self.__path = path

    @classmethod
    def create(cls, path: str):
        return cls(path)

    def write_to(self, writer: List[str]):
        writer.append(f"cmpxchg('{self.__path}')")


class CounterIncludesToken(QueryToken):
    def __init__(self, source_path: str, counter_name: Union[None, str], is_all: bool):
        self.__counter_name = counter_name
        self.__all = is_all
        self.__source_path = source_path

    @classmethod
    def create(cls, source_path: str, counter_name: str) -> CounterIncludesToken:
        return cls(source_path, counter_name, False)

    @classmethod
    def all(cls, source_path: str) -> CounterIncludesToken:
        return cls(source_path, None, True)

    def add_alias_to_path(self, alias: str):
        self.__source_path = alias if not self.__source_path else f"{alias}.{self.__source_path}"

    def write_to(self, writer: List[str]):
        writer.append("counters(")
        if self.__source_path:
            writer.append(self.__source_path)

            if not self.__all:
                writer.append(", ")

        if not self.__all:
            writer.append(f"'{self.__counter_name}'")

        writer.append(")")


class TimeSeriesIncludesToken(QueryToken):
    def __init__(self, source_path: str, time_range: AbstractTimeSeriesRange):
        self.__range = time_range
        self.__source_path = source_path

    @classmethod
    def create(cls, source_path: str, time_range: AbstractTimeSeriesRange) -> TimeSeriesIncludesToken:
        return cls(source_path, time_range)

    def add_alias_to_path(self, alias: str):
        self.__source_path = alias if not self.__source_path else f"{alias}.{self.__source_path}"

    def write_to(self, writer: List[str]):
        writer.append("timeseries(")
        if self.__source_path:
            writer.append(f"{self.__source_path}, ")

        if self.__range.name:
            writer.append(f"'{self.__range.name}', ")

        if isinstance(self.__range, TimeSeriesRange):
            self.__range: TimeSeriesRange

            writer.append(
                f"'{Utils.datetime_to_string(self.__range.from_date)}', " if self.__range.from_date else "null,"
            )
            writer.append(f"'{Utils.datetime_to_string(self.__range.to_date)}', " if self.__range.to_date else "null")

        elif isinstance(self.__range, TimeSeriesTimeRange):
            self.__range: TimeSeriesTimeRange

            if self.__range.type == TimeSeriesRangeType.LAST:
                writer.append("last(")
            else:
                raise ValueError(f"Not supported time range type: {str(self.__range.type)}")
            writer.append(str(self.__range.time.value))
            writer.append(", '")
            writer.append(self.__range.time.unit.value)
            writer.append("')")

        elif isinstance(self.__range, TimeSeriesCountRange):
            self.__range: TimeSeriesCountRange

            if self.__range.type == TimeSeriesRangeType.LAST:
                writer.append("last(")
            else:
                raise ValueError(f"Not supported time range type: {str(self.__range.type)}")

            writer.append(str(self.__range.count))
            writer.append(")")

        else:
            raise ValueError(f"Not supported time range type: {str(self.__range.__class__.__name__)}")

        writer.append(")")


class FieldsToFetchToken(QueryToken):
    def __init__(
        self, fields_to_fetch: List[str], projections: Union[List[str], None], custom_function: bool, source_alias: str
    ):
        self.fields_to_fetch = fields_to_fetch
        self.projections = projections
        self.custom_function = custom_function
        self.source_alias = source_alias

    @classmethod
    def create(
        cls, fields_to_fetch: List[str], projections: List[str], custom_function: bool, source_alias=None
    ) -> FieldsToFetchToken:
        if not fields_to_fetch:
            raise ValueError("fields_to_fetch cannot be None")
        if (not custom_function) and projections is not None and len(projections) != len(fields_to_fetch):
            raise ValueError("Length of projections must be the same as length of field to fetch")

        return FieldsToFetchToken(fields_to_fetch, projections, custom_function, source_alias)

    def write_to(self, writer: List[str]):
        for i in range(len(self.fields_to_fetch)):
            field_to_fetch = self.fields_to_fetch[i]
            if i > 0:
                writer.append(", ")
            if not field_to_fetch:
                writer.append("null")
            else:
                super().write_field(writer, field_to_fetch)

            if self.custom_function:
                continue

            projection = self.projections[i] if self.projections else None
            if not projection or projection == field_to_fetch:
                continue

            writer.append(" as ")
            writer.append(projection)


class DeclareToken(QueryToken):
    def __init__(self, name: str, body: str, parameters: str, time_series: bool):
        self.__name = name
        self.__body = body
        self.__parameters = parameters
        self.__time_series = time_series

    @classmethod
    def create_function(cls, name: str, body: str, parameters: Optional[str] = None) -> DeclareToken:
        return cls(name, body, parameters, False)

    @classmethod
    def create_time_series(cls, name: str, body: str, parameters: Optional[str] = None) -> DeclareToken:
        return cls(name, body, parameters, True)

    def write_to(self, writer: List[str]) -> None:
        writer.append("declare ")
        writer.append("timeseries " if {self.__time_series} else "function ")
        writer.append(self.__name)
        writer.append("(")
        writer.append(self.__parameters)
        writer.append(") ")
        writer.append("{")
        writer.append(os.linesep)
        writer.append(self.__body)
        writer.append(os.linesep)
        writer.append("}")
        writer.append(os.linesep)


class LoadToken(QueryToken):
    def __init__(self, argument: str, alias: str):
        self.argument = argument
        self.alias = alias

    @classmethod
    def create(cls, argument: str, alias: str) -> LoadToken:
        return cls(argument, alias)

    def write_to(self, writer: List[str]) -> None:
        writer.append(self.argument)
        writer.append(" as ")
        writer.append(self.alias)


class FromToken(QueryToken):
    __WHITE_SPACE_CHARS = [" ", "\t", "\r", "\n"]

    def __init__(self, index_name: str, collection_name: str, alias: Optional[str] = None):
        self.__index_name = index_name
        self.__collection_name = collection_name
        self.__dynamic = collection_name is not None
        self.__alias = alias

    @property
    def collection_name(self) -> str:
        return self.__collection_name

    @property
    def index_name(self) -> str:
        return self.__index_name

    @property
    def dynamic(self) -> bool:
        return self.__dynamic

    @property
    def alias(self) -> str:
        return self.__alias

    @classmethod
    def create(cls, index_name: str, collection_name: str, alias: str) -> FromToken:
        return cls(index_name, collection_name, alias)

    def write_to(self, writer: List[str]) -> None:
        if self.__index_name is None and self.__collection_name is None:
            raise ValueError("Either index_name or collection_name must be specified")

        if self.__dynamic:
            writer.append("from '")
            writer.append(Utils.escape_collection_name(self.__collection_name))
            writer.append("'")
        else:
            writer.append("from index '")
            writer.append(self.__index_name)
            writer.append("'")

        if self.__alias is not None:
            writer.append(" as ")
            writer.append(self.__alias)


class OrderByToken(QueryToken):
    @classmethod
    def random(cls) -> OrderByToken:
        return cls("random()", False, OrderingType.STRING)

    @classmethod
    def score_ascending(cls) -> OrderByToken:
        return cls("score()", False, OrderingType.STRING)

    @classmethod
    def score_descending(cls) -> OrderByToken:
        return cls("score()", True, OrderingType.STRING)

    def __init__(self, field_name: str, descending: bool, ordering_or_sorter_name: Union[OrderingType, str]):
        self.__field_name = field_name
        self.__descending = descending

        # duck typing the arg
        is_ordering = isinstance(ordering_or_sorter_name, OrderingType)

        self.__ordering = ordering_or_sorter_name if is_ordering else None
        self.__sorter_name = None if is_ordering else ordering_or_sorter_name

    @classmethod
    def create_distance_ascending_wkt(
        cls, field_name: str, wkt_parameter_name: str, round_factor_parameter_name: str
    ) -> OrderByToken:
        return cls(
            f"spatial.distance({field_name}), "
            f"spatial.wkt(${wkt_parameter_name})"
            f"{'' if round_factor_parameter_name is None else ', $' + round_factor_parameter_name})",
            False,
            OrderingType.STRING,
        )

    @classmethod
    def create_distance_ascending(
        cls,
        field_name: str,
        latitude_parameter_name: str,
        longitude_parameter_name: str,
        round_factor_parameter_name: str,
    ) -> OrderByToken:
        return cls(
            f"spatial.distance({field_name}, "
            f"spatial.point(${latitude_parameter_name}, "
            f"${longitude_parameter_name})"
            f"{'' if round_factor_parameter_name is None else ', $'+round_factor_parameter_name})",
            False,
            OrderingType.STRING,
        )

    @classmethod
    def create_distance_descending_wkt(cls, field_name: str, wkt_parameter_name: str, round_factor_parameter_name: str):
        return cls(
            f"spatial.distance({field_name}, "
            f"spatial.wkt(${wkt_parameter_name})"
            f"{'' if round_factor_parameter_name is None else ', $' + round_factor_parameter_name})",
            True,
            OrderingType.STRING,
        )

    @classmethod
    def create_distance_descending(
        cls,
        field_name: str,
        latitude_parameter_name: str,
        longitude_parameter_name: str,
        round_factor_parameter_name: str,
    ) -> OrderByToken:
        return cls(
            f"spatial.distance({field_name}, "
            f"spatial.point(${latitude_parameter_name}, "
            f"${longitude_parameter_name})"
            f"{'' if round_factor_parameter_name is None else ', $' + round_factor_parameter_name})",
            True,
            OrderingType.STRING,
        )

    @classmethod
    def create_random(cls, seed: str) -> OrderByToken:
        if seed is None:
            raise ValueError("Seed cannot be None")

        return cls("random('" + seed.replace("'", "''") + "')", False, OrderingType.STRING)

    @classmethod
    def create_ascending(cls, field_name: str, sorter_name_or_ordering_type: Union[OrderingType, str]) -> OrderByToken:
        return cls(field_name, False, sorter_name_or_ordering_type)

    @classmethod
    def create_descending(cls, field_name: str, sorter_name_or_ordering_type: Union[OrderingType, str]) -> OrderByToken:
        return cls(field_name, True, sorter_name_or_ordering_type)

    def write_to(self, writer: List[str]) -> None:
        if self.__sorter_name is not None:
            writer.append("custom(")

        self.write_field(writer, self.__field_name)

        if self.__sorter_name is not None:
            writer.append(", '")
            writer.append(self.__sorter_name)
            writer.append("')")
        else:
            if self.__ordering == OrderingType.LONG:
                writer.append(" as long")
            elif self.__ordering == OrderingType.FLOAT:
                writer.append(" as double")
            elif self.__ordering == OrderingType.ALPHA_NUMERIC:
                writer.append(" as alphaNumeric")

        if self.__descending:
            writer.append(" desc")


class GroupByToken(QueryToken):
    def __init__(self, field_name: str, method: GroupByMethod):
        self.__field_name = field_name
        self.__method = method

    @classmethod
    def create(cls, field_name: str, method: Optional[GroupByMethod] = GroupByMethod.NONE) -> GroupByToken:
        return cls(field_name, method)

    def write_to(self, writer: List[str]):
        if self.__method is not GroupByMethod.NONE:
            writer.append("Array(")
        self.write_field(writer, self.__field_name)
        if self.__method is not GroupByMethod.NONE:
            writer.append(")")


class GroupByKeyToken(QueryToken):
    def __init__(self, field_name: str, projected_name: str):
        self.__field_name = field_name
        self.__projected_name = projected_name

    @classmethod
    def create(cls, field_name: str, projected_name: str) -> GroupByKeyToken:
        return cls(field_name, projected_name)

    def write_to(self, writer: List[str]):
        self.write_field(writer, self.__field_name or "key()")
        if self.__projected_name is None or self.__projected_name == self.__field_name:
            return

        writer.append(" as '")
        writer.append(self.__projected_name)
        writer.append("'")


class GroupBySumToken(QueryToken):
    def __init__(self, field_name: str, projected_name: str):
        if field_name is None:
            raise ValueError("field_name cannot be None")

        self.__field_name = field_name
        self.__projected_name = projected_name

    @classmethod
    def create(cls, field_name: str, projected_name: str) -> GroupBySumToken:
        return cls(field_name, projected_name)

    def write_to(self, writer: List[str]):
        writer.append("sum(")
        writer.append(self.__field_name)
        writer.append(")")

        if self.__projected_name is None:
            return

        writer.append(" as ")
        writer.append(self.__projected_name)


class GroupByCountToken(QueryToken):
    def __init__(self, field_name: str):
        self.__field_name = field_name

    @classmethod
    def create(cls, field_name: str) -> GroupByCountToken:
        return cls(field_name)

    def write_to(self, writer: List[str]):
        writer.append("count()")

        if self.__field_name is None:
            return

        writer.append(" as ")
        writer.append(self.__field_name)


class MoreLikeThisToken(QueryToken):
    def __init__(self):
        self.where_tokens: List[QueryToken] = []
        self.document_parameter_name: Union[None, str] = None
        self.options_parameter_name: Union[None, str] = None

    def write_to(self, writer: List[str]):
        writer.append("moreLikeThis(")
        if self.document_parameter_name is None:
            for i in range(len(self.where_tokens)):
                DocumentQueryHelper.add_space_if_needed(
                    self.where_tokens[i - 1] if i > 0 else None, self.where_tokens[i], writer
                )
                self.where_tokens[i].write_to(writer)
        else:
            writer.append("$")
            writer.append(self.document_parameter_name)

        if self.options_parameter_name is None:
            writer.append(")")
            return

        writer.append(", $")
        writer.append(self.options_parameter_name)
        writer.append(")")


class OpenSubclauseToken(QueryToken):
    def __init__(self):
        self.boost_parameter_name: Union[None, str] = None

    @classmethod
    def create(cls) -> OpenSubclauseToken:
        return cls()

    def write_to(self, writer: List[str]):
        if self.boost_parameter_name is not None:
            writer.append("boost")

        writer.append("(")


class CloseSubclauseToken(QueryToken):
    def __init__(self):
        self.boost_parameter_name: Union[None, str] = None

    @classmethod
    def create(cls) -> CloseSubclauseToken:
        return cls()

    def write_to(self, writer: List[str]):
        if self.boost_parameter_name is not None:
            writer.append(", $")
            writer.append(self.boost_parameter_name)

        writer.append(")")


class IntersectMarkerToken(QueryToken):
    __INSTANCE = None

    @classmethod
    def instance(cls) -> IntersectMarkerToken:
        if cls.__INSTANCE is None:
            cls.__INSTANCE = IntersectMarkerToken()
        return cls.__INSTANCE

    def write_to(self, writer: List[str]):
        writer.append(",")


class ShapeToken(QueryToken):
    def __init__(self, shape: str):
        self.__shape = shape

    @classmethod
    def circle(
        cls,
        radius_parameter_name: str,
        latitude_parameter_name: str,
        longitude_parameter_name: str,
        radius_units: SpatialUnits,
    ) -> ShapeToken:
        if radius_units is None:
            return cls(
                f"spatial.circle(${radius_parameter_name}, "
                f"${latitude_parameter_name}, "
                f"${longitude_parameter_name})"
            )

        if radius_units == SpatialUnits.KILOMETERS:
            return cls(
                f"spatial.circle(${radius_parameter_name}, "
                f"${latitude_parameter_name}, "
                f"${longitude_parameter_name}, "
                f"'Kilometers')"
            )

        return cls(
            f"spatial.circle(${radius_parameter_name}, "
            f"${latitude_parameter_name}, "
            f"${longitude_parameter_name}, "
            f"'Miles')"
        )

    @classmethod
    def wkt(cls, shape_wkt_parameter_name: str, units: SpatialUnits) -> ShapeToken:
        if units is None:
            return cls(f"spatial.wkt(${shape_wkt_parameter_name})")

        if units == SpatialUnits.KILOMETERS:
            return cls(f"spatial.wkt(${shape_wkt_parameter_name}, " f"'Kilometers')")

        return cls(f"spatial.wkt(${shape_wkt_parameter_name}, " f"'Miles')")

    def write_to(self, writer: List[str]):
        writer.append(self.__shape)


class WhereToken(QueryToken):
    class MethodsType(enum.Enum):
        CMP_X_CHG = "CmpXChg"

    class WhereMethodCall:
        def __init__(
            self, method_type: WhereToken.MethodsType = None, parameters: List[str] = None, property: str = None
        ):
            self.method_type = method_type
            self.parameters = parameters
            self.property = property

    class WhereOptions:
        def __init__(
            self,
            search: Optional[SearchOperator] = None,
            shape__distance: Optional[Tuple[ShapeToken, float]] = None,
            exact__from__to: Optional[Tuple[bool, Optional[str], Optional[str]]] = None,
            method_type__parameters__property__exact: Optional[
                Tuple[WhereToken.MethodsType, List[str], str, Optional[bool]]
            ] = None,
        ):
            args_count = len(
                list(
                    filter(
                        lambda x: x is not None,
                        [search, shape__distance, exact__from__to, method_type__parameters__property__exact],
                    )
                )
            )

            if args_count > 1:
                raise ValueError(
                    f"WhereOptions init takes only one argument at max. {args_count} arguments were given."
                )

            self.search_operator = None
            self.from_parameter_name = None
            self.to_parameter_name = None
            self.boost = None
            self.fuzzy = None
            self.proximity = None
            self.exact = None
            self.method = None
            self.where_shape = None
            self.distance_error_pct = None

            if search is not None:
                self.search_operator = search

            elif shape__distance is not None:
                self.where_shape = shape__distance[0]
                self.distance_error_pct = shape__distance[1]

            elif exact__from__to is not None:
                self.exact = exact__from__to[0]
                self.from_parameter_name = exact__from__to[1]
                self.to_parameter_name = exact__from__to[2]

            elif method_type__parameters__property__exact is not None:
                self.method = WhereToken.WhereMethodCall(
                    method_type__parameters__property__exact[0],
                    method_type__parameters__property__exact[1],
                    method_type__parameters__property__exact[2],
                )
                self.exact = method_type__parameters__property__exact[3]

        @classmethod
        def default_options(cls) -> WhereToken.WhereOptions:
            return cls()

    def __init__(
        self,
        field_name: str,
        where_operator: WhereOperator,
        parameter_name: str,
        where_options: Optional[WhereToken.WhereOptions] = None,
    ):
        self.field_name = field_name
        self.where_operator = where_operator
        self.parameter_name = parameter_name
        self.options = where_options

    @classmethod
    def create(
        cls,
        op: WhereOperator,
        field_name: str,
        parameter_name: Union[None, str],
        options: Optional[WhereToken.WhereOptions] = None,
    ) -> WhereToken:
        return cls(field_name, op, parameter_name, options or WhereToken.WhereOptions.default_options())

    def add_alias(self, alias: str) -> WhereToken:
        if self.field_name == "id()":
            return self

        return WhereToken(f"{alias}.{self.field_name}", self.where_operator, self.parameter_name, self.options)

    def __write_method(self, writer: List[str]) -> bool:
        if self.options.method is not None:
            if self.options.method.method_type == WhereToken.MethodsType.CMP_X_CHG:
                writer.append("cmpxchg(")
            else:
                raise ValueError(f"Unsupported method: {self.options.method.method_type}")

            first = True
            for parameter in self.options.method.parameters:
                if not first:
                    writer.append(",")
                first = False
                writer.append("$")
                writer.append(parameter)
            writer.append(")")

            if self.options.method.property is not None:
                writer.append("..")
                writer.append(self.options.method.property)

            return True
        return False

    def write_to(self, writer: List[str]) -> None:
        if self.options.boost is not None:
            writer.append("boost(")

        if self.options.fuzzy is not None:
            writer.append("fuzzy(")

        if self.options.proximity is not None:
            writer.append("proximity(")

        if self.options.exact:
            writer.append("exact(")

        if self.where_operator == WhereOperator.SEARCH:
            writer.append("search(")
        elif self.where_operator == WhereOperator.LUCENE:
            writer.append("lucene(")
        elif self.where_operator == WhereOperator.STARTS_WITH:
            writer.append("startsWith(")
        elif self.where_operator == WhereOperator.ENDS_WITH:
            writer.append("endsWith(")
        elif self.where_operator == WhereOperator.EXISTS:
            writer.append("exists(")
        elif self.where_operator == WhereOperator.SPATIAL_WITHIN:
            writer.append("spatial.within(")
        elif self.where_operator == WhereOperator.SPATIAL_CONTAINS:
            writer.append("spatial.contains(")
        elif self.where_operator == WhereOperator.SPATIAL_DISJOINT:
            writer.append("spatial.disjoint(")
        elif self.where_operator == WhereOperator.SPATIAL_INTERSECTS:
            writer.append("spatial.intersects(")
        elif self.where_operator == WhereOperator.REGEX:
            writer.append("regex(")
        elif self.where_operator == WhereOperator.VECTOR_SEARCH:
            writer.append("vector.search(")

        self.__write_inner_where(writer)

        if self.options.exact:
            writer.append(")")

        if self.options.proximity is not None:
            writer.append(", ")
            writer.append(str(self.options.proximity))
            writer.append(")")
        if self.options.fuzzy is not None:
            writer.append(", ")
            writer.append(str(self.options.fuzzy))
            writer.append(")")

        if self.options.boost is not None:
            writer.append(", ")
            writer.append(str(self.options.boost))
            writer.append(")")

    def __write_inner_where(self, writer: List[str]) -> None:
        self.write_field(writer, self.field_name)

        if self.where_operator == WhereOperator.EQUALS:
            writer.append(" = ")
        elif self.where_operator == WhereOperator.NOT_EQUALS:
            writer.append(" != ")
        elif self.where_operator == WhereOperator.GREATER_THAN:
            writer.append(" > ")
        elif self.where_operator == WhereOperator.GREATER_THAN_OR_EQUAL:
            writer.append(" >= ")
        elif self.where_operator == WhereOperator.LESS_THAN:
            writer.append(" < ")
        elif self.where_operator == WhereOperator.LESS_THAN_OR_EQUAL:
            writer.append(" <= ")
        else:
            self.__special_operator(writer)
            return

        if not self.__write_method(writer):
            writer.append("$")
            writer.append(self.parameter_name)

    def __special_operator(self, writer: List[str]):
        if self.where_operator == WhereOperator.IN:
            writer.append(" in ($")
            writer.append(self.parameter_name)
            writer.append(")")
        elif self.where_operator == WhereOperator.ALL_IN:
            writer.append(" all in ($")
            writer.append(self.parameter_name)
            writer.append(")")
        elif self.where_operator == WhereOperator.BETWEEN:
            writer.append(" between $")
            writer.append(self.options.from_parameter_name)
            writer.append(" and $")
            writer.append(self.options.to_parameter_name)
        elif self.where_operator == WhereOperator.SEARCH:
            writer.append(", $")
            writer.append(self.parameter_name)
            if self.options.search_operator == SearchOperator.AND:
                writer.append(", and")
            writer.append(")")
        elif self.where_operator in [
            WhereOperator.LUCENE,
            WhereOperator.STARTS_WITH,
            WhereOperator.ENDS_WITH,
            WhereOperator.REGEX,
        ]:
            writer.append(", $")
            writer.append(self.parameter_name)
            writer.append(")")
        elif self.where_operator == WhereOperator.EXISTS:
            writer.append(")")
        elif self.where_operator in [
            WhereOperator.SPATIAL_WITHIN,
            WhereOperator.SPATIAL_CONTAINS,
            WhereOperator.SPATIAL_DISJOINT,
            WhereOperator.SPATIAL_INTERSECTS,
        ]:
            writer.append(", ")
            self.options.where_shape.write_to(writer)

            if (
                abs(self.options.distance_error_pct - constants.Documents.Indexing.Spatial.DEFAULT_DISTANCE_ERROR_PCT)
                > 1e-40
            ):
                writer.append(", ")
                writer.append(str(self.options.distance_error_pct))
            writer.append(")")

        else:
            raise TypeError("Unexpected WhereToken.where_operator")


class TrueToken(QueryToken):
    __INSTANCE = None

    @classmethod
    def instance(cls) -> TrueToken:
        if cls.__INSTANCE is None:
            cls.__INSTANCE = TrueToken()
        return cls.__INSTANCE

    def write_to(self, writer: List[str]):
        writer.append("true")


class NegateToken(QueryToken):
    __INSTANCE = None

    @classmethod
    def instance(cls) -> NegateToken:
        if cls.__INSTANCE is None:
            cls.__INSTANCE = NegateToken()
        return cls.__INSTANCE

    def write_to(self, writer: List[str]):
        writer.append("not")


class QueryOperatorToken(QueryToken):
    def __init__(self, query_operator: QueryOperator):
        self.__query_operator = query_operator

    @classmethod
    def AND(cls) -> QueryOperatorToken:
        return cls(QueryOperator.AND)

    @classmethod
    def OR(cls) -> QueryOperatorToken:
        return cls(QueryOperator.OR)

    def write_to(self, writer: List[str]):
        if self.__query_operator == QueryOperator.AND:
            writer.append("and")
            return

        writer.append("or")


class HighlightingToken(QueryToken):
    def __init__(self, field_name: str, fragment_length: int, fragment_count: int, operations_parameter_name: str):
        self.__field_name = field_name
        self.__fragment_length = fragment_length
        self.__fragment_count = fragment_count
        self.__operations_parameter_name = operations_parameter_name

    @classmethod
    def create(
        cls, field_name: str, fragment_length: int, fragment_count: int, operations_parameter_name: str
    ) -> HighlightingToken:
        return cls(field_name, fragment_length, fragment_count, operations_parameter_name)

    def write_to(self, writer: List[str]):
        writer.append("highlight(")
        self.write_field(writer, self.__field_name)

        writer.append(",")
        writer.append(str(self.__fragment_length))
        writer.append(",")
        writer.append(str(self.__fragment_count))

        if self.__operations_parameter_name is not None:
            writer.append(",$")
            writer.append(self.__operations_parameter_name)

        writer.append(")")


class ExplanationToken(QueryToken):
    def __init__(self, options_parameter_name: str):
        self.__options_parameter_name = options_parameter_name

    @classmethod
    def create(cls, options_parameter_name: str) -> ExplanationToken:
        return cls(options_parameter_name)

    def write_to(self, writer: List[str]) -> None:
        writer.append("explanations(")
        if self.__options_parameter_name is not None:
            writer.append("$")
            writer.append(self.__options_parameter_name)

        writer.append(")")


class TimingsToken(QueryToken):
    __INSTANCE = None

    @classmethod
    def instance(cls) -> TimingsToken:
        if cls.__INSTANCE is None:
            cls.__INSTANCE = TimingsToken()
        return cls.__INSTANCE

    def write_to(self, writer: List[str]):
        writer.append("timings()")


class DistinctToken(QueryToken):
    __INSTANCE = None

    @classmethod
    def instance(cls) -> DistinctToken:
        if cls.__INSTANCE is None:
            cls.__INSTANCE = DistinctToken()
        return cls.__INSTANCE

    def write_to(self, writer: List[str]):
        writer.append("distinct")


class SuggestToken(QueryToken):
    def __init__(
        self,
        field_name: str,
        alias: Union[None, str],
        term_parameter_name: str,
        options_parameter_name: Union[None, str],
    ):
        if field_name is None:
            raise ValueError("field_name cannot be None")

        if term_parameter_name is None:
            raise ValueError("term_parameter_name cannot be None")

        self.__field_name = field_name
        self.__alias = alias
        self.__term_parameter_name = term_parameter_name
        self.__options_parameter_name = options_parameter_name

    @classmethod
    def create(
        cls,
        field_name: str,
        alias: Union[None, str],
        term_parameter_name: str,
        options_parameter_name: Union[None, str],
    ):
        return cls(field_name, alias, term_parameter_name, options_parameter_name)

    @property
    def field_name(self) -> str:
        return self.__field_name

    def write_to(self, writer: List[str]) -> None:
        writer.append("suggest(")
        writer.append(self.field_name)
        writer.append(", $")
        writer.append(self.__term_parameter_name)

        if self.__options_parameter_name is not None:
            writer.append(", $")
            writer.append(self.__options_parameter_name)

        writer.append(")")

        if not self.__alias or self.__alias.isspace() or self.field_name == self.__alias:
            return

        writer.append(" as ")
        writer.append(self.__alias)


class VectorSearchToken(WhereToken):
    def __init__(
        self,
        wrapped_field_name: str,
        parameter_name: str,
        source_quantization_type: VectorEmbeddingType,
        target_quantization_type: VectorEmbeddingType,
        is_source_base64_encoded: bool,
        is_vector_base64_encoded: bool,
        similarity_threshold: float = None,
        number_of_candidates_for_querying: int = None,
        is_exact: bool = VectorSearch.DEFAULT_IS_EXACT,
    ):
        super().__init__(wrapped_field_name, WhereOperator.VECTOR_SEARCH, parameter_name)
        self._source_quantization_type = source_quantization_type
        self._parameter_name = parameter_name

        self._source_quantization_type = source_quantization_type
        self._target_quantization_type = target_quantization_type

        self._is_source_base64_encoded = is_source_base64_encoded
        self._is_vector_base64_encoded = is_vector_base64_encoded

        self._similarity_threshold = similarity_threshold

        self._number_of_candidates_for_querying = number_of_candidates_for_querying
        self._is_exact = is_exact

    def write_to(self, writer: List[str]) -> None:
        """
        Builds the vector search query string components and appends them to the writer list.
        Follows the same structure as the C# implementation.
        """
        if self._is_exact:
            writer.append("exact(")

        writer.append("vector.search(")

        if (
            self._source_quantization_type == VectorEmbeddingType.SINGLE
            and self._target_quantization_type == VectorEmbeddingType.SINGLE
        ):
            writer.append(self.field_name)
        else:
            method_name = VectorSearch.configuration_to_method_name(
                self._source_quantization_type, self._target_quantization_type
            )
            writer.append(f"{method_name}({self.field_name})")

        # Add main parameter
        writer.append(f", ${self._parameter_name}")

        # Handle optional parameters
        parameters_are_default = self._similarity_threshold is None and self._number_of_candidates_for_querying is None

        if not parameters_are_default:
            # Format similarity threshold with invariant culture
            sim_str = f"{self._similarity_threshold}" if self._similarity_threshold is not None else "null"
            writer.append(f", {sim_str}")

            # Format candidate count
            candidates_str = (
                str(self._number_of_candidates_for_querying)
                if self._number_of_candidates_for_querying is not None
                else "null"
            )
            writer.append(f", {candidates_str}")

        writer.append(")")

        if self._is_exact:
            writer.append(")")



from __future__ import annotations
import enum
import json
from typing import Optional, List, Dict, TYPE_CHECKING, Union, Tuple

import requests

from ravendb.documents.operations.definitions import IOperation
from ravendb.http.http_cache import HttpCache
from ravendb.http.raven_command import RavenCommand, ServerNode
from ravendb.tools.utils import Utils

if TYPE_CHECKING:
    from ravendb.documents.store.definition import DocumentStore
    from ravendb.documents.conventions import DocumentConventions


class CounterOperationType(enum.Enum):
    NONE = "None"
    INCREMENT = "Increment"
    DELETE = "Delete"
    GET = "Get"
    PUT = "Put"


class CounterOperation:
    def __init__(
        self,
        counter_name: str,
        counter_operation_type: CounterOperationType,
        delta: Optional[int] = None,
    ):
        if not counter_name:
            raise ValueError("Missing counter_name property")
        if not counter_operation_type:
            raise ValueError(f"Missing counter_operation_type property in counter {counter_name}")
        if delta is None and counter_operation_type == CounterOperationType.INCREMENT:
            raise ValueError(f"Missing delta property in counter {counter_name} of type {counter_operation_type}")
        self.counter_name = counter_name
        self.delta = delta
        self.counter_operation_type = counter_operation_type
        self._change_vector = None
        self._document_id = None

    @classmethod
    def create(cls, counter_name: str, counter_operation_type: CounterOperationType, delta: Optional[int] = None):
        return cls(counter_name, counter_operation_type, delta)

    def to_json(self):
        return {
            "Type": self.counter_operation_type.value,
            "CounterName": self.counter_name,
            "Delta": self.delta,
        }


class DocumentCountersOperation:
    def __init__(self, document_id: str, operations: List[CounterOperation]):
        if not document_id:
            raise ValueError("Missing document_id property")

        self._document_id = document_id
        self._operations = operations

    def add_operations(self, operation: CounterOperation):
        if self._operations is None:
            self._operations = []

        self._operations.append(operation)

    @property
    def operations(self):
        return self._operations

    def to_json(self):
        if not self._operations:
            raise ValueError("Missing operations property on Counters")
        return {
            "DocumentId": self._document_id,
            "Operations": [operation.to_json() for operation in self._operations],
        }


class CounterBatch:
    def __init__(
        self,
        reply_with_all_nodes_values: Optional[bool] = None,
        documents: Optional[List[DocumentCountersOperation]] = None,
        from_etl: Optional[bool] = None,
    ):
        self.reply_with_all_nodes_values = reply_with_all_nodes_values
        self.documents = documents
        self.from_etl = from_etl

    def to_json(self) -> Dict:
        return {
            "ReplyWithAllNodesValues": self.reply_with_all_nodes_values,
            "Documents": [document_counters_operation.to_json() for document_counters_operation in self.documents],
            "FromEtl": self.from_etl,
        }


class CounterDetail:
    def __init__(
        self,
        document_id: Optional[str] = None,
        counter_name: Optional[str] = None,
        total_value: Optional[int] = None,
        etag: Optional[int] = None,
        counter_values: Optional[Dict[str, int]] = None,
    ):
        self.document_id = document_id
        self.counter_name = counter_name
        self.total_value = total_value
        self.etag = etag
        self.counter_values = counter_values

    def to_json(self) -> Dict:
        return {
            "DocumentId": self.document_id,
            "CounterName": self.counter_name,
            "TotalValue": self.total_value,
            "Etag": self.etag,
            "CounterValues": self.counter_values,
        }

    @classmethod
    def from_json(cls, json_dict: Dict) -> Optional[CounterDetail]:
        return (
            cls(
                json_dict["DocumentId"],
                json_dict["CounterName"],
                json_dict["TotalValue"],
                json_dict.get("Etag", None),
                json_dict["CounterValues"],
            )
            if json_dict is not None
            else None
        )


class CountersDetail:
    def __init__(self, counters: List[CounterDetail]):
        self.counters = counters

    def to_json(self) -> Dict:
        return {"Counters": [counter.to_json() for counter in self.counters]}

    @classmethod
    def from_json(cls, json_dict: Dict) -> CountersDetail:
        return cls([CounterDetail.from_json(counter_detail_json) for counter_detail_json in json_dict["Counters"]])


class CounterBatchOperation(IOperation[CountersDetail]):
    def __init__(self, counter_batch: CounterBatch):
        self._counter_batch = counter_batch

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> RavenCommand[CountersDetail]:
        return CounterBatchOperation.CounterBatchCommand(self._counter_batch)

    class CounterBatchCommand(RavenCommand[CountersDetail]):
        def __init__(self, counter_batch: CounterBatch):
            super().__init__(CountersDetail)

            if counter_batch is None:
                raise ValueError("Counter batch cannot be None")

            self._counter_batch = counter_batch

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "POST", f"{node.url}/databases/{node.database}/counters", data=self._counter_batch.to_json()
            )

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                return

            self.result = CountersDetail.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return False


class GetCountersOperation(IOperation[CountersDetail]):
    def __init__(
        self, doc_id: str, counters: Optional[Union[str, List[str]]] = None, return_full_results: Optional[bool] = None
    ):
        self._doc_id = doc_id
        self._counters = [] if counters is None else counters if isinstance(counters, list) else [counters]
        self._return_full_results = return_full_results

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> RavenCommand[CounterDetail]:
        return GetCountersOperation.GetCounterValuesCommand(
            self._doc_id, self._counters, self._return_full_results, conventions
        )

    class GetCounterValuesCommand(RavenCommand[CountersDetail]):
        def __init__(
            self, doc_id: str, counters: List[str], return_full_results: bool, conventions: DocumentConventions
        ):
            super(GetCountersOperation.GetCounterValuesCommand, self).__init__(CountersDetail)

            if doc_id is None:
                raise ValueError("Doc id cannot be None")

            self._doc_id = doc_id
            self._counters = counters
            self._return_full_results = return_full_results
            self._conventions = conventions

        def create_request(self, node: ServerNode) -> requests.Request:
            path_builder = [node.url]
            path_builder.append("/databases/")
            path_builder.append(node.database)
            path_builder.append("/counters?docId=")
            path_builder.append(Utils.quote_key(self._doc_id))

            request = requests.Request("GET")

            if self._counters:
                if len(self._counters) > 1:
                    request = self._prepare_request_with_multiple_counters(path_builder, request)
                else:
                    path_builder.append("&counter=")
                    path_builder.append(Utils.quote_key(self._counters[0]))

            if self._return_full_results and request.method == "GET":
                path_builder.append("&full=true")

            request.url = "".join(path_builder)

            return request

        def _prepare_request_with_multiple_counters(
            self, path_builder: List[str], request: requests.Request
        ) -> requests.Request:
            unique_names, sum_length = self._get_ordered_unique_names()

            # if it is too big, we drop to POST (note that means that we can't use the HTTP cache any longer)
            # we are fine with that, such requests are going to be rare
            if sum_length < 1024:
                for unique_name in unique_names:
                    path_builder.append("&counter=")
                    path_builder.append(Utils.quote_key(unique_name or ""))
            else:
                request = requests.Request("POST")

                doc_ops = DocumentCountersOperation(self._doc_id, [])

                for counter in unique_names:
                    counter_operation = CounterOperation(counter, CounterOperationType.GET)
                    doc_ops.operations.append(counter_operation)

                batch = CounterBatch()
                batch.documents = [doc_ops]
                batch.reply_with_all_nodes_values = self._return_full_results

                request.data = batch.to_json()
            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                return

            self.result = CountersDetail.from_json(json.loads(response))

        def _get_ordered_unique_names(self) -> Tuple[List[str], int]:
            unique_names = set(self._counters)
            sum_length = sum([len(counter) if counter else 0 for counter in unique_names])
            return list(unique_names), sum_length

        def is_read_request(self) -> bool:
            return True



from __future__ import annotations

import datetime
import json
from http import HTTPStatus
from typing import Union, List, Generic, TypeVar, Type, Callable, Dict, TYPE_CHECKING, Optional

from ravendb.documents.operations.compare_exchange.compare_exchange_value_result_parser import (
    CompareExchangeValueResultParser,
)

from ravendb.primitives import constants
from ravendb.documents.queries.facets.misc import FacetResult
from ravendb.documents.queries.index_query import IndexQuery
from ravendb.documents.session.document_info import DocumentInfo
from ravendb.documents.session.loaders.loaders import LazyMultiLoaderWithInclude
from ravendb.documents.store.lazy import Lazy
from ravendb.documents.session.conditional_load import ConditionalLoadResult
from ravendb.exceptions.raven_exceptions import RavenException
from ravendb.tools.utils import Utils, CaseInsensitiveDict
from ravendb.documents.queries.query import QueryResult
from ravendb.extensions.json_extensions import JsonExtensions
from ravendb.documents.commands.crud import GetDocumentsResult, ConditionalGetResult
from ravendb.documents.session.operations.load_operation import LoadOperation
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.operations.lazy.definition import LazyOperation
from ravendb.documents.commands.multi_get import Content, GetRequest, GetResponse

if TYPE_CHECKING:
    from ravendb.documents.session.operations.query import QueryOperation
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )
    from ravendb.documents.session.document_session import DocumentSession
    from ravendb.documents.session.cluster_transaction_operation import (
        ClusterTransactionOperationsBase,
    )

_T = TypeVar("_T")


class IndexQueryContent(Content):
    def __init__(self, conventions: DocumentConventions, query: IndexQuery):
        self.__conventions = conventions
        self.__query = query

    def write_content(self) -> dict:
        return JsonExtensions.write_index_query(self.__conventions, self.__query)


class LazyStartsWithOperation(LazyOperation):
    def __init__(
        self,
        object_type: Type[_T],
        prefix: str,
        matches: str,
        exclude: str,
        start: int,
        page_size: int,
        session_operations: "InMemoryDocumentSessionOperations",
        start_after: str,
    ):
        self.__object_type = object_type
        self.__prefix = prefix
        self.__matches = matches
        self.__exclude = exclude
        self.__start = start
        self.__page_size = page_size
        self.__session_operations = session_operations
        self.__start_after = start_after

        self.result = None
        self.query_result: Union[None, QueryResult] = None
        self._requires_retry: Union[None, bool] = None

    @property
    def requires_retry(self) -> bool:
        return self._requires_retry

    def create_request(self) -> GetRequest:
        request = GetRequest()
        request.url = "/docs"
        request.query = (
            f"?startsWith={Utils.quote_key(self.__prefix)}"
            f"&matches={Utils.quote_key(self.__matches or '')}"
            f"&exclude={Utils.quote_key(self.__exclude or '')}"
            f"&start={self.__start}&pageSize={self.__page_size}"
            f"&startAfter={self.__start_after or ''}"
        )
        return request

    def handle_response(self, response: GetResponse) -> None:
        get_documents_result = GetDocumentsResult.from_json(json.loads(response.result))
        final_results = CaseInsensitiveDict()

        for document in get_documents_result.results:
            new_document_info = DocumentInfo.get_new_document_info(document)
            self.__session_operations._documents_by_id.add(new_document_info)

            if new_document_info.key is None:
                continue  # todo: is this possible?

            if self.__session_operations.is_deleted(new_document_info.key):
                final_results[new_document_info.key] = None
                continue

            doc = self.__session_operations._documents_by_id.get_value(new_document_info.key)
            if doc is not None:
                final_results[new_document_info.key] = self.__session_operations.track_entity_document_info(
                    self.__object_type, doc
                )
                continue

            final_results[new_document_info.key] = None

        self.result = final_results


class LazyConditionalLoadOperation(LazyOperation):
    def __init__(
        self, object_type: Type[_T], key: str, change_vector: str, session: "InMemoryDocumentSessionOperations"
    ):
        self.__object_type = object_type
        self.__key = key
        self.__change_vector = change_vector
        self.__session = session
        self.__result: Union[None, object] = None
        self.__requires_retry: Union[None, bool] = None

    @property
    def result(self) -> object:
        return self.__result

    @property
    def query_result(self) -> QueryResult:
        raise NotImplementedError()

    @property
    def requires_retry(self) -> bool:
        return self.__requires_retry

    def create_request(self) -> GetRequest:
        request = GetRequest()

        request.url = "/docs"
        request.method = "GET"
        request.query = f"?id={Utils.quote_key(self.__key)}"

        request.headers[constants.Headers.IF_NONE_MATCH] = f'"{self.__change_vector}"'
        return request

    def handle_response(self, response: "GetResponse") -> None:
        if response.force_retry:
            self.__result = None
            self.__requires_retry = None
            return

        if response.status_code == HTTPStatus.NOT_MODIFIED:
            self.__result = ConditionalLoadResult.create(None, self.__change_vector)
            return

        elif response.status_code == HTTPStatus.NOT_FOUND:
            self.__session.register_missing(self.__key)
            self.__result = ConditionalLoadResult.create(None, None)
            return

        try:
            if response.result is not None:
                etag = response.headers.get(constants.Headers.ETAG)

                res = ConditionalGetResult.from_json(json.loads(response.result))
                document_info = DocumentInfo.get_new_document_info(res.results[0])
                r = self.__session.track_entity_document_info(self.__object_type, document_info)

                self.__result = ConditionalLoadResult.create(r, etag)
                return

            self.__result = None
            self.__session.register_missing(self.__key)
        except Exception as e:
            raise RuntimeError(e)


class LazySessionOperations:
    def __init__(self, delegate: DocumentSession):
        self._session = delegate

    def include(self, path: str) -> LazyMultiLoaderWithInclude:
        return LazyMultiLoaderWithInclude(self._session).include(path)

    def load(
        self, ids: Union[List[str], str], object_type: Optional[Type[_T]] = None, on_eval: Callable = None
    ) -> Optional[Lazy[Union[Dict[str, object], object]]]:
        if not ids:
            return None

        if isinstance(ids, str):
            key = ids
            if self._session.advanced.is_loaded(key):
                return Lazy(lambda: self._session.load(key, object_type))

            lazy_load_operation = LazyLoadOperation(
                object_type, self._session, LoadOperation(self._session).by_key(key)
            ).by_key(key)

            return self._session.add_lazy_operation(object_type, lazy_load_operation, on_eval)

        elif isinstance(ids, list):
            return self._session.lazy_load_internal(object_type, ids, [], on_eval)

        raise TypeError(f"Expected a 'str' or 'list' of 'str', the document ids. Got '{type(ids).__name__}'.")

    def load_starting_with(
        self,
        id_prefix: str,
        object_type: Optional[Type[_T]] = None,
        matches: str = None,
        start: int = 0,
        page_size: int = 25,
        exclude: str = None,
        start_after: str = None,
    ) -> Lazy[Dict[str, _T]]:
        operation = LazyStartsWithOperation(
            object_type, id_prefix, matches, exclude, start, page_size, self._session, start_after
        )

        return self._session.add_lazy_operation(dict, operation, None)

    def conditional_load(
        self, key: str, change_vector: str, object_type: Optional[Type[_T]] = None
    ) -> Lazy[ConditionalLoadResult[_T]]:
        if not key or key.isspace():
            raise ValueError("key cannot be None or whitespace")

        if self._session.is_loaded(key):

            def __lazy_factory():
                entity = self._session.load(key, object_type)
                if entity is None:
                    return ConditionalLoadResult.create(None, None)
                cv = self._session.advanced.get_change_vector_for(entity)
                return ConditionalLoadResult.create(entity, cv)

            return Lazy(__lazy_factory)

        if change_vector.isspace():
            raise ValueError(
                f"The requested document with id '{key}' is not loaded into the session and could not "
                f"conditional load when change_vector is None or empty"
            )

        lazy_load_operation = LazyConditionalLoadOperation(object_type, key, change_vector, self._session)
        return self._session.add_lazy_operation(ConditionalLoadResult, lazy_load_operation, None)


class LazyLoadOperation(LazyOperation):
    def __init__(
        self, object_type: Type[_T], session: "InMemoryDocumentSessionOperations", load_operation: LoadOperation
    ):
        self.__object_type = object_type
        self.__session = session
        self.__load_operation = load_operation
        self.__keys: Union[None, List[str]] = None
        self.__includes: Union[None, List[str]] = None
        self.__already_in_session: List[str] = []

        self.__result: Union[None, List[_T]] = None
        self.__query_result: Union[None, QueryResult] = None
        self.__requires_retry: Union[None, bool] = None

    @property
    def query_result(self) -> QueryResult:
        return self.__query_result

    @query_result.setter
    def query_result(self, value) -> None:
        self.__query_result = value

    @property
    def result(self) -> List[_T]:
        return self.__result

    @result.setter
    def result(self, value) -> None:
        self.__result = value

    @property
    def requires_retry(self) -> bool:
        return self.__requires_retry

    def create_request(self) -> Optional[GetRequest]:
        query_builder = ["?"]
        if self.__includes is not None:
            for include in self.__includes:
                query_builder.append(f"&include={include}")

        has_items = False

        for key in self.__keys:
            if self.__session.is_loaded_or_deleted(key):
                self.__already_in_session.append(key)
            else:
                has_items = True
                query_builder.append(f"&id={Utils.escape(key,None,None)}")

        if not has_items:
            self.result = self.__load_operation.get_documents(self.__object_type)
            return None

        get_request = GetRequest()
        get_request.url = "/docs"
        get_request.query = "".join(query_builder)
        return get_request

    def by_key(self, key: str) -> LazyLoadOperation:
        if not key or key.isspace():
            return self

        if self.__keys is None:
            self.__keys = [key]

        return self

    def by_keys(self, keys: List[str]) -> LazyLoadOperation:
        self.__keys = list(set(filter(lambda key: key, keys)))
        return self

    def with_includes(self, *includes: str) -> LazyLoadOperation:
        self.__includes = includes
        return self

    def handle_response(self, response: GetResponse) -> None:
        if response.force_retry:
            self.result = None
            self.requires_retry = True
            return
        json_result = json.loads(response.result)
        multi_load_result = None if json_result is None else GetDocumentsResult.from_json(json_result)
        self.__handle_response(multi_load_result)

    def __handle_response(self, load_result: GetDocumentsResult) -> None:
        if self.__already_in_session:
            # push this to the session
            LoadOperation(self.__session).by_keys(self.__already_in_session).get_documents(self.__object_type)

        self.__load_operation.set_result(load_result)
        if not self.requires_retry:
            self.result = self.__load_operation.get_documents(self.__object_type)


class LazyQueryOperation(Generic[_T], LazyOperation[_T]):
    def __init__(
        self,
        object_type: Type[_T],
        session: "InMemoryDocumentSessionOperations",
        query_operation: QueryOperation,
        after_query_executed: List[Callable[[QueryResult], None]],
    ):
        self.__object_type = object_type
        self.__session = session
        self.__query_operation = query_operation
        self.__after_query_executed = after_query_executed

        self.__result: Union[None, List[_T]] = None
        self.__query_result: Union[None, QueryResult] = None
        self.__requires_retry: Union[None, bool] = None

    def create_request(self) -> "GetRequest":
        request = GetRequest()
        request.can_cache_aggressively = (
            not self.__query_operation.index_query.disable_caching
            and not self.__query_operation.index_query.wait_for_non_stale_results
        )
        request.url = "/queries"
        request.method = "POST"
        request.query = f"?queryHash={self.__query_operation.index_query.get_query_hash()}"
        request.content = IndexQueryContent(self.__session.conventions, self.__query_operation.index_query)
        return request

    @property
    def result(self) -> Union[None, List[_T]]:
        return self.__result

    @result.setter
    def result(self, value: List[_T]):
        self.__result = value

    @property
    def query_result(self) -> QueryResult:
        return self.__query_result

    @query_result.setter
    def query_result(self, value: QueryResult):
        self.__query_result = value

    @property
    def requires_retry(self) -> bool:
        return self.__requires_retry

    def handle_response(self, response: "GetResponse") -> None:
        if response.force_retry:
            self.result = None
            self.__requires_retry = True
            return

        query_result = None

        if response.result is not None:
            json_response = json.loads(response.result)
            query_result = None if json_response is None else QueryResult.from_json(json_response)

        self.__handle_response(query_result, response.elapsed)

    def __handle_response(self, query_result: QueryResult, duration: datetime.timedelta):
        self.__query_operation.ensure_is_acceptable_and_save_result(query_result, duration)

        for event in self.__after_query_executed:
            event(query_result)

        self.__result = self.__query_operation.complete(self.__object_type)
        self.__query_result = query_result


class LazyAggregationQueryOperation(LazyOperation):
    def __init__(
        self,
        session: "InMemoryDocumentSessionOperations",
        index_query: IndexQuery,
        invoke_after_query_executed: Callable[[QueryResult], None],
        process_results: Callable[[QueryResult], Dict[str, FacetResult]],
    ):
        self.__session = session
        self.__index_query = index_query
        self.__invoke_after_query_executed = invoke_after_query_executed
        self.__process_results = process_results

        self.result: Union[None, object] = None
        self.query_result: Union[None, QueryResult] = None
        self.__requires_retry: Union[None, bool] = None

    def create_request(self) -> "GetRequest":
        request = GetRequest()
        request.url = "/queries"
        request.method = "POST"
        request.query = f"?queryHash={self.__index_query.get_query_hash()}"
        request.content = IndexQueryContent(self.__session.conventions, self.__index_query)
        return request

    def handle_response(self, response: "GetResponse") -> None:
        if response.force_retry:
            self.result = None
            self.__requires_retry = True
            return

        query_result = QueryResult.from_json(json.loads(response.result))
        self.__handle_response(query_result)

    def __handle_response(self, query_result: QueryResult) -> None:
        self.result = self.__process_results(query_result)
        self.query_result = query_result

    @property
    def requires_retry(self) -> bool:
        return self.__requires_retry


class LazySuggestionQueryOperation(LazyOperation):
    def __init__(
        self,
        session: "InMemoryDocumentSessionOperations",
        index_query: IndexQuery,
        invoke_after_query_executed: Callable[[QueryResult], None],
        process_results: Callable[[QueryResult], Dict[str, FacetResult]],
    ):
        self.__session = session
        self.__index_query = index_query
        self.__invoke_after_query_executed = invoke_after_query_executed
        self.__process_results = process_results

        self.result: Union[None, object] = None
        self.query_result: Union[None, QueryResult] = None
        self.__requires_retry: Union[None, bool] = None

    def create_request(self) -> "GetRequest":
        request = GetRequest()
        request.can_cache_aggressively = (
            not self.__index_query.disable_caching and self.__index_query.wait_for_non_stale_results
        )
        request.url = "/queries"
        request.method = "POST"
        request.query = f"?queryHash={self.__index_query.get_query_hash()}"
        request.content = IndexQueryContent(self.__session.conventions, self.__index_query)
        return request

    def handle_response(self, response: "GetResponse") -> None:
        if response.force_retry:
            self.result = None
            self.__requires_retry = True
            return

        query_result = QueryResult.from_json(json.loads(response.result))
        self.__handle_response(query_result)

    def __handle_response(self, query_result: QueryResult) -> None:
        self.result = self.__process_results(query_result)
        self.query_result = query_result

    @property
    def requires_retry(self) -> bool:
        return self.__requires_retry


class LazyGetCompareExchangeValueOperation(Generic[_T], LazyOperation[_T]):
    def __init__(
        self,
        cluster_session: ClusterTransactionOperationsBase,
        key: str,
        conventions: DocumentConventions,
        object_type: Type[_T] = None,
    ):
        if cluster_session is None:
            raise ValueError("Cluster session cannot be None")
        if conventions is None:
            raise ValueError("Conventions cannot be None")
        if key is None:
            raise ValueError("Key cannot be None")

        self.__cluster_session = cluster_session
        self.__object_type = object_type
        self.__conventions = conventions
        self.__key = key

        self.__result = None
        self.__requires_retry = None

    @property
    def result(self) -> object:
        return self.__result

    @property
    def query_result(self) -> QueryResult:
        raise NotImplementedError("Not implemented")

    @property
    def requires_retry(self) -> bool:
        return self.__requires_retry

    def create_request(self) -> Optional["GetRequest"]:
        if self.__cluster_session.is_tracked(self.__key):
            self.__result, not_tracked = self.__cluster_session.get_compare_exchange_value_from_session_internal(
                self.__key, self.__object_type
            )
            return None

        request = GetRequest()
        request.url = "/cmpxchg"
        request.method = "GET"
        query_builder = f"?key={Utils.quote_key(self.__key)}"
        request.query = query_builder

        return request

    def handle_response(self, response: "GetResponse") -> None:
        if response.force_retry:
            self.__result = None
            self.__requires_retry = True
            return

        try:
            if response.result is not None:
                value = CompareExchangeValueResultParser.get_value(dict, response.result, False, self.__conventions)

                if self.__cluster_session.session.no_tracking:
                    if value is None:
                        self.__result = self.__cluster_session.register_missing_compare_exchange_value(
                            self.__key
                        ).get_value(self.__object_type, self.__conventions)
                        return

                    self.__result = self.__cluster_session.register_compare_exchange_value(value).get_value(
                        self.__object_type, self.__conventions
                    )
                    return

                if value is not None:
                    self.__cluster_session.register_compare_exchange_value(value)

            if not self.__cluster_session.is_tracked(self.__key):
                self.__cluster_session.register_missing_compare_exchange_value(self.__key)

            self.__result, not_tracked = self.__cluster_session.get_compare_exchange_value_from_session_internal(
                self.__key, self.__object_type
            )
        except Exception as e:
            raise RavenException(f"Unable to get compare exchange value: {self.__key}", e)


class LazyGetCompareExchangeValuesOperation(LazyOperation[_T], Generic[_T]):
    def __init__(
        self,
        cluster_session: ClusterTransactionOperationsBase,
        keys: List[str],
        conventions: DocumentConventions,
        object_type: Type[_T] = None,
        starts_with: str = None,
        start: int = 0,
        page_size: int = 0,
    ):
        if not keys:
            raise ValueError("Keys cannot be None or empty")

        if cluster_session is None:
            raise ValueError("Cluster session cannot be None")

        if conventions is None:
            raise ValueError("Conventions cannot be None")

        self.__cluster_session = cluster_session
        self.__keys = keys
        self.__conventions = conventions
        self.__object_type = object_type

        self.__start = start
        self.__page_size = page_size
        self.__starts_with = starts_with

        self.__result: Optional[object] = None
        self.__requires_retry: Optional[bool] = None

    @property
    def result(self) -> object:
        return self.__result

    @property
    def query_result(self) -> QueryResult:
        raise NotImplementedError("Not implemented.")

    @property
    def requires_retry(self) -> bool:
        return self.__requires_retry

    def create_request(self) -> Optional["GetRequest"]:
        path_builder = None

        if self.__keys:
            for key in self.__keys:
                if self.__cluster_session.is_tracked(key):
                    continue

                if path_builder is None:
                    path_builder = ["?"]

                path_builder.append("&key=")
                path_builder.append(Utils.quote_key(key))

        else:
            path_builder = ["?"]
            if self.__starts_with and not self.__starts_with.isspace():
                path_builder.append("&startsWith")
                path_builder.append(Utils.quote_key(self.__starts_with))
            path_builder.append("&start=")
            path_builder.append(self.__start)
            path_builder.append("&pageSize=")
            path_builder.append(self.__page_size)

        if path_builder is None:
            self.__result, _ = self.__cluster_session.get_compare_exchange_values_from_session_internal(
                self.__keys, self.__object_type
            )
            return None

        request = GetRequest()
        request.url = "/cmpxchg"
        request.method = "GET"
        request.query = "".join(path_builder)

        return request

    def handle_response(self, response: "GetResponse") -> None:
        if response.force_retry:
            self.__result = None
            self.__requires_retry = True
            return

        try:
            if response.result is not None:
                if self.__cluster_session.session.no_tracking:
                    result = CaseInsensitiveDict()
                    for key, value in CompareExchangeValueResultParser.get_values(
                        dict, response.result, False, self.__conventions
                    ).items():
                        if value is None:
                            result[key] = self.__cluster_session.register_missing_compare_exchange_value(key).get_value(
                                self.__object_type, self.__conventions
                            )
                            continue

                        result[key] = self.__cluster_session.register_compare_exchange_value(value).get_value(
                            self.__object_type, self.__conventions
                        )
                    self.__result = result
                    return

                for key, value in CompareExchangeValueResultParser.get_values(
                    dict, response.result, False, self.__conventions
                ).items():
                    if value is None:
                        continue

                    self.__cluster_session.register_compare_exchange_value(value)

            if self.__keys is not None:
                for key in self.__keys:
                    if self.__cluster_session.is_tracked(key):
                        continue

                    self.__cluster_session.register_missing_compare_exchange_value(key)

            self.__result, _ = self.__cluster_session.get_compare_exchange_values_from_session_internal(
                self.__keys, self.__object_type
            )

        except Exception:
            raise RavenException(f"Unable to get compare exchange values:  {' ,'.join(self.__keys)}")



import datetime
import json
from typing import TYPE_CHECKING, Union, Optional

import requests

from ravendb.documents.operations.definitions import MaintenanceOperation, OperationIdResult, IOperation
from ravendb.documents.queries.index_query import IndexQuery
from ravendb.http.http_cache import HttpCache
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions
    from ravendb.documents import DocumentStore


class GetOperationStateOperation(MaintenanceOperation):
    def __init__(self, key: int, node_tag: str = None):
        self.__key = key
        self.__node_tag = node_tag

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand:
        return self.GetOperationStateCommand(self.__key, self.__node_tag)

    class GetOperationStateCommand(RavenCommand[dict]):
        def __init__(self, key: int, node_tag: str = None):
            super().__init__(dict)
            self.__key = key
            self.__selected_node_tag = node_tag

            self.__timeout = datetime.timedelta(seconds=15)

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("GET", f"{node.url}/databases/{node.database}/operations/state?id={self.__key}")

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return
            self.result = json.loads(response)


class QueryOperationOptions(object):
    def __init__(
        self,
        allow_stale: bool = True,
        stale_timeout: datetime.timedelta = None,
        max_ops_per_sec: int = None,
        retrieve_details: bool = False,
    ):
        self.allow_stale = allow_stale
        self.stale_timeout = stale_timeout
        self.retrieve_details = retrieve_details
        self.max_ops_per_sec = max_ops_per_sec


class DeleteByQueryOperation(IOperation[OperationIdResult]):
    def __init__(self, query_to_delete: Union[str, IndexQuery], options: Optional[QueryOperationOptions] = None):
        if not query_to_delete:
            raise ValueError("Invalid query")
        if isinstance(query_to_delete, str):
            query_to_delete = IndexQuery(query=query_to_delete)
        self.__query_to_delete = query_to_delete
        self.__options = options if options is not None else QueryOperationOptions()

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: Optional[HttpCache] = None
    ) -> RavenCommand[OperationIdResult]:
        return self.__DeleteByIndexCommand(conventions, self.__query_to_delete, self.__options)

    class __DeleteByIndexCommand(RavenCommand[OperationIdResult]):
        def __init__(
            self,
            conventions: "DocumentConventions",
            query_to_delete: IndexQuery,
            options: Optional[QueryOperationOptions] = None,
        ):
            super().__init__(OperationIdResult)
            self.__conventions = conventions
            self.__query_to_delete = query_to_delete
            self.__options = options if options is not None else QueryOperationOptions()

        def create_request(self, server_node):
            url = server_node.url + "/databases/" + server_node.database + "/queries"
            path = (
                f"?allowStale={self.__options.allow_stale}&maxOpsPerSec={self.__options.max_ops_per_sec or ''}"
                f"&details={self.__options.retrieve_details}"
            )
            if self.__options.stale_timeout is not None:
                path += "&staleTimeout=" + str(self.__options.stale_timeout)

            url += path
            data = self.__query_to_delete.to_json()
            return requests.Request("DELETE", url, data=data)

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()
            response = json.loads(response)
            self.result = OperationIdResult(response["OperationId"], response["OperationNodeTag"])

        def is_read_request(self) -> bool:
            return False



# todo: DatabaseChanges class
import time
from typing import Callable, TYPE_CHECKING, Optional

from ravendb.documents.operations.definitions import OperationExceptionResult
from ravendb.exceptions.exception_dispatcher import ExceptionDispatcher
from ravendb.http.raven_command import RavenCommand
from ravendb.primitives.exceptions import OperationCancelledException
from ravendb.tools.utils import Utils
from ravendb.documents.operations.misc import GetOperationStateOperation

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions
    from ravendb.http.request_executor import RequestExecutor
    from ravendb.changes.database_changes import DatabaseChanges


class Operation:
    def __init__(
        self,
        request_executor: "RequestExecutor",
        changes: Optional[Callable[[], "DatabaseChanges"]],
        conventions: "DocumentConventions",
        key: int,
        node_tag: str = None,
    ):
        self.__request_executor = request_executor
        self.__conventions = conventions
        self.__key = key
        self.node_tag = node_tag

    def fetch_operations_status(self) -> dict:
        command = self._get_operation_state_command(self.__conventions, self.__key, self.node_tag)
        self.__request_executor.execute_command(command)

        return command.result

    def _get_operation_state_command(
        self, conventions: "DocumentConventions", key: int, node_tag: str = None
    ) -> RavenCommand[dict]:
        return GetOperationStateOperation.GetOperationStateCommand(self.__key, node_tag)

    def wait_for_completion(self) -> None:
        while True:
            status = self.fetch_operations_status()
            operation_status = status.get("Status")

            if operation_status == "Completed":
                return
            elif operation_status == "Canceled":
                raise OperationCancelledException()
            elif operation_status == "Faulted":
                result = status.get("Result")
                exception_result: OperationExceptionResult = Utils.initialize_object(
                    result, OperationExceptionResult, True
                )
                schema = ExceptionDispatcher.ExceptionSchema(
                    self.__request_executor.url, exception_result.type, exception_result.message, exception_result.error
                )
                raise ExceptionDispatcher.get(schema, exception_result.status_code)

            time.sleep(0.5)



import uuid
from typing import Optional


class RaftIdGenerator:
    @staticmethod
    def new_id() -> str:
        return str(uuid.uuid4())

    @staticmethod
    def dont_care_id() -> str:
        return ""


class StartingWithOptions:
    def __init__(self, start_with: str, start: Optional[int] = None, page_size: Optional[int] = None):
        self.starts_with = start_with
        self.start = start
        self.page_size = page_size



from __future__ import annotations

import json
from datetime import datetime
from typing import List, Optional, TYPE_CHECKING

import requests

from ravendb.http.raven_command import RavenCommand
from ravendb.json.result import JsonArrayResult
from ravendb.tools.utils import Utils

if TYPE_CHECKING:
    from ravendb.http.server_node import ServerNode


class GetRevisionsCommand(RavenCommand[JsonArrayResult]):
    def __init__(
        self,
        id_: str = None,
        start: int = None,
        page_size: int = None,
        metadata_only: bool = None,
        before: datetime = None,
        change_vector: str = None,
        change_vectors: List[str] = None,
    ):
        super().__init__(JsonArrayResult)
        self._id = id_
        self._start = start
        self._page_size = page_size
        self._metadata_only = metadata_only
        self._before = before
        self._change_vector = change_vector
        self._change_vectors = change_vectors

    @property
    def change_vectors(self) -> Optional[List[str]]:
        return self._change_vectors

    @property
    def change_vector(self) -> Optional[str]:
        return self._change_vector

    @property
    def before(self) -> Optional[datetime]:
        return self._before

    @property
    def id(self) -> Optional[str]:
        return self._id

    @classmethod
    def from_change_vector(cls, change_vector: str, metadata_only: bool = False) -> GetRevisionsCommand:
        return GetRevisionsCommand(change_vector=change_vector, metadata_only=metadata_only)

    @classmethod
    def from_change_vectors(cls, change_vectors: List[str] = None, metadata_only: bool = False) -> GetRevisionsCommand:
        return GetRevisionsCommand(change_vectors=change_vectors, metadata_only=metadata_only)

    @classmethod
    def from_before(cls, id_: str, before: datetime) -> GetRevisionsCommand:
        if id_ is None:
            raise ValueError("Id cannot be None")
        return GetRevisionsCommand(id_=id_, before=before)

    @classmethod
    def from_start_page(
        cls, id_: str = None, start: int = None, page_size: int = None, metadata_only: bool = False
    ) -> GetRevisionsCommand:
        if id_ is None:
            raise ValueError("Id cannot be None")

        return cls(id_=id_, start=start, page_size=page_size, metadata_only=metadata_only)

    def create_request(self, node: ServerNode) -> requests.Request:
        path_builder = [node.url, "/databases/", node.database, "/revisions?"]

        self.get_request_query_string(path_builder)
        return requests.Request("GET", "".join(path_builder))

    def get_request_query_string(self, path_builder: List[str]) -> None:
        if self._id is not None:
            path_builder.append("&id=")
            path_builder.append(Utils.escape(self._id))
        elif self._change_vector is not None:
            path_builder.append("&changeVector=")
            path_builder.append(Utils.escape(self._change_vector))
        elif self._change_vectors is not None:
            for change_vector in self._change_vectors:
                path_builder.append("&changeVector=")
                path_builder.append(Utils.escape(change_vector))

        if self._before is not None:
            path_builder.append("&before")
            path_builder.append(Utils.datetime_to_string(self._before))

        if self._start is not None:
            path_builder.append("&start=")
            path_builder.append(str(self._start))

        if self._page_size is not None:
            path_builder.append("&pageSize=")
            path_builder.append(str(self._page_size))

        if self._metadata_only:
            path_builder.append("&metadataOnly=true")

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if response is None:
            self.result = None
            return

        self.result = JsonArrayResult.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return True


class GetRevisionsBinEntryCommand(RavenCommand[JsonArrayResult]):
    def __init__(self, etag: int, page_size: int):
        super().__init__(JsonArrayResult)
        self._etag = etag
        self._page_size = page_size

    def create_request(self, node: ServerNode) -> requests.Request:
        request = requests.Request("GET")
        path = [node.url, "/databases/", node.database, "/revisions/bin?start=", str(self._etag)]
        if self._page_size is not None:
            path.append("&pageSize=")
            path.append(str(self._page_size))

        request.url = "".join(path)

        return request

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if response is None:
            self._throw_invalid_response()

        self.result = JsonArrayResult.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return True
